# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains a markdown story exporter."""

from __future__ import unicode_literals

import tabulate

from timesketch.lib.stories import interface
from timesketch.lib.stories import manager


class MarkdownStoryExporter(interface.StoryExporter):
    """Markdown story exporter."""

    # String representing the output format of the story.
    EXPORT_FORMAT = 'markdown'

    # Number of rows of a DataFrame to include at the top of a markdown table.
    _DATAFRAM_HEADER_ROWS = 20

    # Number of rows of a DataFrame to include as a trailer of a table.
    _DATAFRAM_TAIL_ROWS = 5

    # The number of rows
    def _dataframe_to_markdown(self, data_frame):
        """Returns a markdown formatted string from a pandas DataFrame."""
        nr_rows, _ = data_frame.shape
        if not nr_rows:
            return '*<empty table>*'

        if nr_rows <= (self._DATAFRAM_HEADER_ROWS + self._DATAFRAM_TAIL_ROWS):
            return tabulate.tabulate(
                data_frame, tablefmt='pipe', headers='keys')

        return_lines = []
        return_lines.append(tabulate.tabulate(
            data_frame[:self._DATAFRAM_HEADER_ROWS], tablefmt='pipe',
            headers='keys'))
        return_lines.append('| ... |')
        return_lines.append(tabulate.tabulate(
            data_frame[-self._DATAFRAM_TAIL_ROWS:], tablefmt='pipe',
            headers='keys'))
        return '\n'.join(return_lines)

    def export_story(self):
        """Export the story as a markdown."""
        return_strings = []
        for line_dict in self._data_lines:
            line_type = line_dict.get('type', '')
            if line_type == 'text':
                return_strings.append(line_dict.get('value', ''))

            elif line_type == 'aggregation':
                aggregation_data = line_dict.get('value')
                aggregation = aggregation_data.get('aggregation')
                if not aggregation:
                    return_strings.append(
                        '**Unable to fetch aggregation data**')
                    continue
                return_strings.append(
                    self._dataframe_to_markdown(aggregation.to_pandas()))

            elif line_type == 'dataframe':
                return_strings.append(
                    self._dataframe_to_markdown(line_dict.get('value')))

            elif line_type == 'chart':
                return_strings.append(
                    '*<unable_to_display_chart_objects>*')

        return '\n\n'.join(return_strings)


manager.StoryExportManager.register_exporter(MarkdownStoryExporter)



# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains utilities for analyzers."""

from __future__ import unicode_literals

from six.moves import urllib_parse as urlparse

import numpy

from timesketch.lib.analyzers import interface


# Title and header text of a story that is common among browser
# based analyzers.
BROWSER_STORY_TITLE = 'Browser Artifacts'
BROWSER_STORY_HEADER = """
This is an automatically generated story that browser history
based analyzers contribute to. Each section in this story
is generated by a separate analyzer.
"""

# Title and header text of a story that is common among browser
# based analyzers.
SIGMA_STORY_TITLE = 'Sigma Artifacts'
SIGMA_STORY_HEADER = """
This is an automatically generated story that Sigma
based analyzers contribute to.
"""

# CDN domain list based on:
# https://github.com/WPO-Foundation/webpagetest/blob/master/agent/wpthook/cdn.h
# Last updated: 2019-01-11
KNOWN_CDN_DOMAINS = {
    '.att-dsa.net': 'AT&T',
    '.pix-cdn.org': 'Advanced Hosters CDN',
    '.akamai.net': 'Akamai',
    '.akamaiedge.net': 'Akamai',
    '.akamaihd.net': 'Akamai',
    '.akamaitechnologies.com': 'Akamai',
    '.akamaitechnologies.fr': 'Akamai',
    '.akamaized.net': 'Akamai',
    '.edgekey.net': 'Akamai',
    '.edgesuite.net': 'Akamai',
    '.srip.net': 'Akamai',
    '.tl88.net': 'Akamai China CDN',
    '.gslb.tbcache.com': 'Alimama',
    '.cloudfront.net': 'Amazon CloudFront',
    '.aads-cn.net': 'Aryaka',
    '.aads-cng.net': 'Aryaka',
    '.aads1.net': 'Aryaka',
    '.azion.net': 'Azion',
    '.azioncdn.com': 'Azion',
    '.azioncdn.net': 'Azion',
    '.bo.lt': 'BO.LT',
    '.bisongrid.net': 'Bison Grid',
    '.bitgravity.com': 'BitGravity',
    '.bluehatnetwork.com': 'Blue Hat Network',
    '.b-cdn.net': 'BunnyCDN',
    '.cdn77.net': 'CDN77',
    '.cdn77.org': 'CDN77',
    '.cdngc.net': 'CDNetworks',
    '.gccdn.net': 'CDNetworks',
    '.panthercdn.com': 'CDNetworks',
    '.cdnsun.net': 'CDNsun',
    '.cdnvideo.net': 'CDNvideo',
    '.cdnvideo.ru': 'CDNvideo',
    '.cachefly.net': 'Cachefly',
    '.caspowa.com': 'Caspowa',
    '.cedexis.net': 'Cedexis',
    '.ccgslb.com': 'ChinaCache',
    '.lxdns.com': 'ChinaNetCenter',
    '.ourwebpic.com': 'ChinaNetCenter',
    '.wscdns.com': 'ChinaNetCenter',
    '.wscloudcdn.com': 'ChinaNetCenter',
    '.cloudflare.com': 'Cloudflare',
    '.cotcdn.net': 'Cotendo CDN',
    '.systemcdn.net': 'Edgecast',
    '.transactcdn.net': 'Edgecast',
    '.v1cdn.net': 'Edgecast',
    '.v2cdn.net': 'Edgecast',
    '.v3cdn.net': 'Edgecast',
    '.v4cdn.net': 'Edgecast',
    '.v5cdn.net': 'Edgecast',
    '.edgecastcdn.net': 'Edgecast',
    '.cdninstagram.com': 'Facebook',
    '.fbcdn.net': 'Facebook',
    '.fastly.net': 'Fastly',
    '.fastlylb.net': 'Fastly',
    '.nocookie.net': 'Fastly',
    '.cdn.gocache.net': 'GoCache',
    '.doubleclick.net': 'Google',
    '.googleusercontent.com': 'Google',
    '.gstatic.com': 'Google',
    '.googlehosted.com': 'Google',
    '.googlesyndication.': 'Google',
    '.hiberniacdn.com': 'HiberniaCDN',
    '.hwcdn.net': 'Highwinds',
    '.hosting4cdn.com': 'Hosting4CDN',
    '.incapdns.net': 'Incapsula',
    '.inscname.net': 'Instart Logic',
    '.insnw.net': 'Instart Logic',
    '.internapcdn.net': 'Internap',
    '.kinxcdn.com': 'KINX CDN',
    '.kinxcdn.net': 'KINX CDN',
    '.kxcdn.com': 'KeyCDN',
    '.lswcdn.eu': 'LeaseWeb CDN',
    '.lswcdn.net': 'LeaseWeb CDN',
    '.footprint.net': 'Level 3',
    '.fpbns.net': 'Level 3',
    '.llnwd.net': 'Limelight',
    '.cdncloud.net.au': 'MediaCloud',
    '.mncdn.com': 'Medianova',
    '.mncdn.net': 'Medianova',
    '.mncdn.org': 'Medianova',
    '.azure.microsoft.com': 'Microsoft Azure',
    '.azureedge.net': 'Microsoft Azure',
    '.vo.msecnd.net': 'Microsoft Azure',
    '.instacontent.net': 'Mirror Image',
    '.mirror-image.net': 'Mirror Image',
    '.ngenix.net': 'NGENIX',
    '.nyiftw.com': 'NYI FTW',
    '.nyiftw.net': 'NYI FTW',
    '.netdna-cdn.com': 'NetDNA',
    '.netdna-ssl.com': 'NetDNA',
    '.netdna.com': 'NetDNA',
    '.netlify.com': 'Netlify',
    '.r.worldcdn.net': 'OnApp',
    '.r.worldssl.net': 'OnApp',
    '.optimalcdn.com': 'Optimal CDN',
    '.pagerain.net': 'PageRain',
    '.raxcdn.com': 'Rackspace',
    '.resrc.it': 'ReSRC.it',
    '.rlcdn.com': 'Reapleaf',
    '.rncdn1.com': 'Reflected Networks',
    '.rncdn7.com': 'Reflected Networks',
    '.revcn.net': 'Rev Software',
    '.revdn.net': 'Rev Software',
    '.roast.io': 'Roast.io',
    '.streamprovider.net': 'Rocket CDN',
    '.cdn.sfr.net': 'SFR',
    '.simplecdn.net': 'Simple CDN',
    '.singularcdn.net.br': 'Singular CDN',
    '.stackpathdns.com': 'StackPath',
    '.swiftcdn1.com': 'SwiftCDN',
    '.swiftserve.com': 'SwiftCDN',
    '.trbcdn.ru': 'TRBCDN',
    '.gslb.taobao.com': 'Taobao',
    '.taobaocdn.com': 'Taobao',
    '.tbcdn.cn': 'Taobao',
    '.cdntel.net': 'Telenor',
    '.twimg.com': 'Twitter',
    '.unicorncdn.net': 'UnicornCDN',
    '.voxcdn.net': 'VoxCDN',
    '.gravatar.com': 'WordPress',
    '.wordpress.com': 'WordPress',
    '.wp.com': 'WordPress',
    '.ay1.b.yahoo.com': 'Yahoo',
    '.yahooapis.com': 'Yahoo',
    '.yimg.': 'Yahoo',
    '.yottaa.net': 'Yottaa',
    '.zenedge.net': 'Zenedge',
    '.afxcdn.net': 'afxcdn.net',
    '.cubecdn.net': 'cubeCDN',
    '.cdn.jsdelivr.net': 'jsDelivr',
    '.squixa.net': 'section.io'}


def get_domain_from_url(url):
    """Extract domain from URL.

    Args:
        url: URL to parse.

    Returns:
        String with domain from URL.
    """
    # TODO: See if we can optimize this because it is rather slow.
    domain_parsed = urlparse.urlparse(url)
    domain_full = domain_parsed.netloc
    domain, _, _ = domain_full.partition(':')
    return domain


def get_tld_from_domain(domain):
    """Get the top level domain from a domain string.

    Args:
        domain: string with a full domain, eg. www.google.com

    Returns:
        string: TLD or a top level domain extracted from the domain,
        eg: google.com
    """
    return '.'.join(domain.split('.')[-2:])


def strip_www_from_domain(domain):
    """Strip www. from beginning of domain names.

    Args:
        domain: string with a full domain, eg. www.google.com

    Returns:
        string: Domain without any www, eg: google.com
    """
    if domain.startswith('www.'):
        return domain[4:]
    return domain


def get_cdn_provider(domain):
    """Return name of CDN provider if domain is recognized as a CDN.

    Args:
        domain: Domain name to check against CDN list.

    Returns:
        String of names of CDN providers or empty string if not found.

    """
    cdn_providers = [v for k, v in iter(KNOWN_CDN_DOMAINS.items()) if
                     domain.endswith(k.lower())]
    return ' '.join(set(cdn_providers))


def _fix_np_nan(source_dict, attribute, replace_with=None):
    """Replaces a numpy.nan value within dict with another value.

    Args:
        source_dict: a dictionary.
        attribute: string that is a key to the dictionary, in which
            numpy.nan values will be replaced by the supplied value.
        replace_with: the value that will be used as a replacement for
            the numpy.nan. If not supplied an empty list will be used.
    """
    if replace_with is None:
        replace_with = []

    value = source_dict.get(attribute)
    try:
        if numpy.isnan(value):
            source_dict[attribute] = replace_with
    except TypeError:
        # The value does not need to be changed.
        pass


def get_events_from_data_frame(frame, datastore):
    """Generates events from a data frame.

    Args:
        frame: a pandas DataFrame object.
        datastore: Elasticsearch datastore client.

    Yields:
        An event (interface.Event) object for each row
        in the DataFrame.
    """
    for row in frame.iterrows():
        _, entry = row
        event_id = entry.get('_id')
        if not event_id:
            continue
        event_index = entry.get('_index')
        if not event_index:
            continue
        event_type = entry.get('_type')

        source = entry.to_dict()
        _fix_np_nan(source, '__ts_emojis')
        _fix_np_nan(source, 'human_readable', replace_with='')
        _fix_np_nan(source, 'tag')

        datetime = source.get('datetime')
        if hasattr(datetime, 'to_pydatetime'):
            datetime_string = datetime.to_pydatetime().isoformat()
            source['datetime'] = datetime_string

        event_dict = dict(
            _id=event_id, _type=event_type, _index=event_index,
            _source=source)
        yield interface.Event(event_dict, datastore)



"""Tests for FeatureExtractionPlugin."""
from __future__ import unicode_literals

import os
import re
import yaml

import mock

from timesketch.lib import emojis
from timesketch.lib.analyzers import feature_extraction
from timesketch.lib.testlib import BaseTest

from timesketch.lib.testlib import MockDataStore

class TestFeatureExtractionPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    def _config_validation(self, config):
        """Validate that all items of a config are valid."""
        query = config.get('query_string', config.get('query_dsl'))
        self.assertIsNotNone(query)
        self.assertIsInstance(query, str)

        attribute = config.get('attribute')
        self.assertIsNotNone(attribute)

        store_as = config.get('store_as')
        self.assertIsNotNone(store_as)

        expression = config.get('re')
        self.assertIsNotNone(expression)
        try:
            _ = re.compile(expression)
        except re.error as exception:
            self.assertIsNone(exception)

        emojis_to_add = config.get('emojis')
        if emojis_to_add:
            self.assertIsInstance(emojis_to_add, (list, tuple))
            for emoji_name in emojis_to_add:
                emoji_code = emojis.get_emoji(emoji_name)
                self.assertNotEqual(emoji_code, '')

        tags = config.get('tags')
        if tags:
            self.assertIsInstance(tags, (list, tuple))

        create_view = config.get('create_view')
        if create_view:
            self.assertIsInstance(create_view, bool)

        aggregate = config.get('aggregate')
        if aggregate:
            self.assertIsInstance(aggregate, bool)

    # TODO: Add tests for the feature extraction.
    def test_config(self):
        """Tests that the config file is valid."""
        config_file = os.path.join('data', 'features.yaml')
        self.assertTrue(os.path.isfile(config_file))

        with open(config_file) as fh:
            config = yaml.safe_load(fh)

        self.assertIsInstance(config, dict)

        for key, value in iter(config.items()):
            self.assertIsInstance(key, str)
            self.assertIsInstance(value, dict)
            self._config_validation(value)

    # Mock the Elasticsearch datastore.
    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def test_get_attribute_value(self):
        """Test function _get_attribute_value()."""
        analyzer = feature_extraction.FeatureExtractionSketchPlugin(
            'test_index', 1)
        current_val = ['hello']
        extracted_value = ['hello']
        # pylint: disable=protected-access
        new_val = analyzer._get_attribute_value(
            current_val=current_val,
            extracted_value=extracted_value,
            keep_multi=True,
            merge_values=True,
            type_list=True)
        new_val.sort()

        self.assertEqual(new_val, ['hello'])

        current_val = ['hello']
        extracted_value = ['hello2', 'hello3']
        # pylint: disable=protected-access
        new_val = analyzer._get_attribute_value(
            current_val,
            extracted_value,
            True,
            True,
            True)
        new_val.sort()

        self.assertEqual(new_val, ['hello', 'hello2', 'hello3'])

        current_val = ['hello']
        extracted_value = ['hello2', 'hello3']
        # pylint: disable=protected-access
        new_val = analyzer._get_attribute_value(
            current_val,
            extracted_value,
            False,
            True,
            True)
        new_val.sort()

        self.assertEqual(new_val, ['hello', 'hello2'])

        current_val = ['hello']
        extracted_value = ['hello2', 'hello3']
        # pylint: disable=protected-access
        new_val = analyzer._get_attribute_value(
            current_val,
            extracted_value,
            False,
            False,
            True)
        new_val.sort()

        self.assertEqual(new_val, ['hello2'])

        current_val = ['hello']
        extracted_value = ['hello2', 'hello3']
        # pylint: disable=protected-access
        new_val = analyzer._get_attribute_value(
            current_val,
            extracted_value,
            True,
            False,
            True)
        new_val.sort()

        self.assertEqual(new_val, ['hello2', 'hello3'])

        current_val = 'hello'
        extracted_value = ['hello2', 'hello3']
        # pylint: disable=protected-access
        new_val = analyzer._get_attribute_value(
            current_val,
            extracted_value,
            True,
            True,
            False)

        self.assertEqual(new_val, 'hello,hello2,hello3')

        current_val = 'hello'
        extracted_value = ['hello2', 'hello3']
        # pylint: disable=protected-access
        new_val = analyzer._get_attribute_value(
            current_val,
            extracted_value,
            False,
            True,
            False)

        self.assertEqual(new_val, 'hello,hello2')

        current_val = 'hello'
        extracted_value = ['hello2', 'hello3']
        # pylint: disable=protected-access
        new_val = analyzer._get_attribute_value(
            current_val,
            extracted_value,
            True,
            False,
            False)

        self.assertEqual(new_val, 'hello2,hello3')

        current_val = 'hello'
        extracted_value = ['hello2', 'hello3']
        # pylint: disable=protected-access
        new_val = analyzer._get_attribute_value(
            current_val,
            extracted_value,
            False,
            False,
            False)

        self.assertEqual(new_val, 'hello2')



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for end-to-end tests."""

import time
import unittest
import os
import inspect
import traceback
import collections

from timesketch_api_client import client as api_client
from timesketch_import_client import importer

# Default values based on Docker config.
TEST_DATA_DIR = '/usr/local/src/timesketch/end_to_end_tests/test_data'
HOST_URI = 'http://127.0.0.1'
USERNAME = 'test'
PASSWORD = 'test'


class BaseEndToEndTest(object):
    """Base class for end to end tests.

    Attributes:
        api: Instance of an API client
        sketch: Instance of Sketch object
        assertions: Instance of unittest.TestCase
    """

    NAME = 'name'

    def __init__(self):
        """Initialize the end-to-end test object."""
        self.api = api_client.TimesketchApi(
            host_uri=HOST_URI, username=USERNAME, password=PASSWORD)
        self.sketch = self.api.create_sketch(name=self.NAME)
        self.assertions = unittest.TestCase()
        self._counter = collections.Counter()

    def import_timeline(self, filename):
        """Import a Plaso, CSV or JSONL file.

        Args:
            filename (str): Filename of the file to be imported.

        Raises:
            TimeoutError if import takes too long.
        """
        file_path = os.path.join(TEST_DATA_DIR, filename)
        print('Importing: {0:s}'.format(file_path))

        with importer.ImportStreamer() as streamer:
            streamer.set_sketch(self.sketch)
            streamer.set_timeline_name(file_path)
            streamer.add_file(file_path)
            timeline = streamer.timeline

        # Poll the timeline status and wait for the timeline to be ready
        max_time_seconds = 600  # Timeout after 10min
        sleep_time_seconds = 5  # Sleep between API calls
        max_retries = max_time_seconds / sleep_time_seconds
        retry_count = 0
        while True:
            if retry_count >= max_retries:
                raise TimeoutError
            _ = timeline.lazyload_data(refresh_cache=True)
            status = timeline.status
            # TODO: Do something with other statuses? (e.g. failed)
            if status == 'ready':
                break
            retry_count += 1
            time.sleep(sleep_time_seconds)

    def _get_test_methods(self):
        """Inspect class and list all methods that matches the criteria.

        Yields:
            Function name and bound method.
        """
        for name, func in inspect.getmembers(self, predicate=inspect.ismethod):
            if name.startswith('test_'):
                yield name, func

    def setup(self):
        """Setup function that is run before any tests.

        This is a good place to import any data that is needed.
        """
        return NotImplementedError

    def run_tests(self):
        """Run all test functions from the class.

        Returns:
            Counter of number of tests and errors.
        """
        print('*** {0:s} ***'.format(self.NAME))
        for test_name, test_func in self._get_test_methods():
            self._counter['tests'] += 1
            print('Running test: {0:s} ...'.format(
                test_name), end='', flush=True)
            try:
                test_func()
            except Exception:  # pylint: disable=broad-except
                # TODO: Change to logging module instead of prints
                print(traceback.format_exc())
                self._counter['errors'] += 1
                continue
            print('[OK]')
        return self._counter



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph plugin for Chrome downloads."""

from timesketch.lib.graphs.interface import BaseGraphPlugin
from timesketch.lib.graphs import manager


class ChromeDownloadsGraph(BaseGraphPlugin):
    """Graph plugin for Chrome downloads and executions.

    This plugin will try to match any downloaded files with Windows prefetch
    executions.
    """

    NAME = 'ChromeDownloads'
    DISPLAY_NAME = 'Chrome downloads'

    def generate(self):
        """Generate the graph.

        Returns:
            Graph object instance.
        """
        query = 'data_type:"chrome:history:file_downloaded"'
        return_fields = [
            'hostname', 'received_bytes', 'full_path', 'url', 'domain'
        ]

        chrome_events = self.event_stream(
            query_string=query, return_fields=return_fields)

        for chrome_event in chrome_events:
            computer_name = chrome_event['_source'].get('hostname', '')
            full_path = chrome_event['_source'].get('full_path', '')
            received_bytes = chrome_event['_source'].get('received_bytes', '')
            url = chrome_event['_source'].get('url', '')
            domain = chrome_event['_source'].get('domain', '')

            # Note: If any component of the path contains a \, it will be split
            # split, potentially wrong split. No data will be lost, but the
            # file node will have part of the path as the filename.
            # TODO: Revisit this and see if we can make it more robust.
            if '\\' in full_path:
                separator = '\\'
            else:
                separator = '/'

            filename = full_path.split(separator)[-1]
            computer = self.graph.add_node(computer_name, {'type': 'computer'})

            # If full patch is missing, set url as label.
            if not full_path:
                full_path = url
                filename = f'UNKNOWN file download from {domain}'

            file = self.graph.add_node(filename, {
                'full_path': full_path,
                'received_bytes': received_bytes,
                'type': 'file'
            })

            edge_label = f'Downloaded {received_bytes} bytes'
            self.graph.add_edge(computer, file, edge_label, chrome_event)

            # Get prefetch events that matches the filename.
            prefetch_query = f'executable:"*{filename}"'
            prefetch_return_fields = ['executable', 'hostname']
            prefetch_events = self.event_stream(
                query_string=prefetch_query,
                return_fields=prefetch_return_fields,
                scroll=False
            )
            for prefetch_event in prefetch_events:
                computer_name = prefetch_event['_source'].get('hostname')
                executable_name = prefetch_event['_source'].get('executable')
                computer = self.graph.add_node(
                    computer_name, {'type': 'computer'})
                executable = self.graph.add_node(executable_name, {
                    'type': 'file'
                })
                self.graph.add_edge(
                    computer, executable, 'Executed', prefetch_event)

        self.graph.commit()

        return self.graph


manager.GraphManager.register_graph(ChromeDownloadsGraph)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Story resources for version 1 of the Timesketch API."""

from flask import jsonify
from flask import request
from flask import abort
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user
from sqlalchemy import desc

from timesketch.api.v1 import resources
from timesketch.lib import forms
from timesketch.lib.definitions import HTTP_STATUS_CODE_CREATED
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.lib.stories import api_fetcher as story_api_fetcher
from timesketch.lib.stories import manager as story_export_manager
from timesketch.models import db_session
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import Story


class StoryListResource(resources.ResourceMixin, Resource):
    """Resource to get all stories for a sketch or to create a new story."""

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            Stories in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')

        stories = []
        for story in Story.query.filter_by(
                sketch=sketch).order_by(desc(Story.created_at)):
            stories.append(story)
        return self.to_json(stories)

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            A view in JSON (instance of flask.wrappers.Response)
        """
        form = forms.StoryForm.build(request)
        if not form.validate_on_submit():
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST, 'Unable to validate form data.')

        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'write'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have write access controls on sketch.')

        title = ''
        if form.title.data:
            title = form.title.data
        story = Story(
            title=title, content='[]', sketch=sketch, user=current_user)
        db_session.add(story)
        db_session.commit()
        return self.to_json(story, status_code=HTTP_STATUS_CODE_CREATED)


class StoryResource(resources.ResourceMixin, Resource):
    """Resource to get a story."""

    @staticmethod
    def _export_story(story, sketch_id, export_format='markdown'):
        """Returns a story in a format as requested in export_format.

        Args:
            story: a story object (instance of Story) that is to be exported.
            sketch_id: integer with the sketch ID.
            export_format: string with the name of the format to export the
                story to. Defaults to "markdown".

        Returns:
            The exported story in the format described. This could be a text
            or a binary, depending on the output format.
        """
        exporter_class = story_export_manager.StoryExportManager.get_exporter(
            export_format)
        if not exporter_class:
            return b''

        with exporter_class() as exporter:
            data_fetcher = story_api_fetcher.ApiDataFetcher()
            data_fetcher.set_sketch_id(sketch_id)

            exporter.set_data_fetcher(data_fetcher)
            exporter.set_title(story.title)
            exporter.set_creation_date(story.created_at.isoformat())
            if story.user:
                exporter.set_author(story.user.username)
            exporter.set_exporter(current_user.username)

            exporter.from_string(story.content)
            return exporter.export_story()

    @login_required
    def get(self, sketch_id, story_id):
        """Handles GET request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model
            story_id: Integer primary key for a story database model

        Returns:
            A story in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        story = Story.query.get(story_id)

        if not story:
            msg = 'No Story found with this ID.'
            abort(HTTP_STATUS_CODE_NOT_FOUND, msg)

        if not sketch:
            msg = 'No sketch found with this ID.'
            abort(HTTP_STATUS_CODE_NOT_FOUND, msg)

        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')

        # Check that this story belongs to the sketch
        if story.sketch_id != sketch.id:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND,
                'Sketch ID ({0:d}) does not match with the ID in '
                'the story ({1:d})'.format(sketch.id, story.sketch_id))

        # Only allow editing if the current user is the author.
        # This is needed until we have proper collaborative editing and
        # locking implemented.
        meta = dict(is_editable=False)
        if current_user == story.user:
            meta['is_editable'] = True

        return self.to_json(story, meta=meta)

    @login_required
    def post(self, sketch_id, story_id):
        """Handles POST request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model
            story_id: Integer primary key for a story database model

        Returns:
            A view in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        story = Story.query.get(story_id)

        if not story:
            msg = 'No Story found with this ID.'
            abort(HTTP_STATUS_CODE_NOT_FOUND, msg)

        if not sketch:
            msg = 'No sketch found with this ID.'
            abort(HTTP_STATUS_CODE_NOT_FOUND, msg)

        if story.sketch_id != sketch.id:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND,
                'Sketch ID ({0:d}) does not match with the ID in '
                'the story ({1:d})'.format(sketch.id, story.sketch_id))

        if not sketch.has_permission(current_user, 'write'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have write access controls on sketch.')

        form = request.json
        if not form:
            form = request.data

        if form and form.get('export_format'):
            export_format = form.get('export_format')
            return jsonify(story=self._export_story(
                story=story, sketch_id=sketch_id, export_format=export_format))

        story.title = form.get('title', '')
        story.content = form.get('content', '[]')
        db_session.add(story)
        db_session.commit()
        return self.to_json(story, status_code=HTTP_STATUS_CODE_CREATED)

    @login_required
    def delete(self, sketch_id, story_id):
        """Handles DELETE request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model
            story_id: Integer primary key for a story database model
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        story = Story.query.get(story_id)

        if not story:
            msg = 'No Story found with this ID.'
            abort(HTTP_STATUS_CODE_NOT_FOUND, msg)

        if not sketch:
            msg = 'No sketch found with this ID.'
            abort(HTTP_STATUS_CODE_NOT_FOUND, msg)

        # Check that this timeline belongs to the sketch
        if story.sketch_id != sketch.id:
            msg = (
                'The sketch ID ({0:d}) does not match with the story'
                'sketch ID ({1:d})'.format(sketch.id, story.sketch_id))
            abort(HTTP_STATUS_CODE_FORBIDDEN, msg)

        if not sketch.has_permission(user=current_user, permission='write'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'The user does not have write permission on the sketch.')

        sketch.stories.remove(story)
        db_session.commit()
        return HTTP_STATUS_CODE_OK



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Table chart."""
import altair as alt

from timesketch.lib.charts import manager
from timesketch.lib.charts import interface


class TableChart(interface.BaseChart):
    """Table chart object."""

    NAME = 'table'

    def generate(self):
        """Generate the chart.

        Returns:
            Instance of altair.Chart
        """
        # Only the data attribute is used for table rendering but we need the
        # mark to construct the chart.
        chart = alt.Chart().mark_bar()
        chart.data = self.values
        return chart


manager.ChartManager.register_chart(TableChart)



# Copyright 2018 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Analyzer module."""

# Register all analyzers here by importing them.
from timesketch.lib.analyzers import account_finder
from timesketch.lib.analyzers import browser_search
from timesketch.lib.analyzers import browser_timeframe
from timesketch.lib.analyzers import chain
from timesketch.lib.analyzers import domain
from timesketch.lib.analyzers import expert_sessionizers
from timesketch.lib.analyzers import feature_extraction
from timesketch.lib.analyzers import login
from timesketch.lib.analyzers import phishy_domains
from timesketch.lib.analyzers import safebrowsing
from timesketch.lib.analyzers import sessionizer
from timesketch.lib.analyzers import sigma_tagger
from timesketch.lib.analyzers import similarity_scorer
from timesketch.lib.analyzers import ssh_sessionizer
from timesketch.lib.analyzers import gcp_servicekey
from timesketch.lib.analyzers import ntfs_timestomp
from timesketch.lib.analyzers import yetiindicators
from timesketch.lib.analyzers import win_crash
from timesketch.lib.analyzers import win_evtxgap
from timesketch.lib.analyzers import tagger



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module holds archive API calls for version 1 of the Timesketch API."""

from __future__ import unicode_literals

import datetime
import io
import json
import logging
import zipfile

from flask import abort
from flask import current_app
from flask import jsonify
from flask import request
from flask import send_file
from flask_login import current_user
from flask_login import login_required
from flask_restful import Resource

import pandas as pd

from timesketch import version
from timesketch.api.v1 import export
from timesketch.api.v1 import resources
from timesketch.api.v1 import utils
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK
from timesketch.lib.stories import manager as story_export_manager
from timesketch.models.sketch import Event
from timesketch.models.sketch import Sketch


logger = logging.getLogger('timesketch.api_archive')


class SketchArchiveResource(resources.ResourceMixin, Resource):
    """Resource to archive a sketch."""

    DEFAULT_QUERY_FILTER = {
        'size': 10000,
        'terminate_after': 10000
    }

    def __init__(self, **kwargs):
        super(SketchArchiveResource, self).__init__(**kwargs)
        self._sketch = None
        self._sketch_indices = None

    @property
    def sketch_indices(self):
        """Returns a set of sketch indices that are ready."""
        if not self._sketch_indices:
            if not self._sketch:
                return set()

            self._sketch_indices = {
                t.searchindex.index_name
                for t in self._sketch.timelines
                if t.get_status.status.lower() == 'ready'
            }
        return self._sketch_indices

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.

        Returns:
            A sketch in JSON (instance of flask.wrappers.Response)
        """
        if current_user.admin:
            sketch = Sketch.query.get(sketch_id)
        else:
            sketch = Sketch.query.get_with_acl(sketch_id)

        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'read'):
            if not current_user.admin:
                abort(
                    HTTP_STATUS_CODE_FORBIDDEN, (
                        'User does not have sufficient access rights to '
                        'read the sketch.'))

        meta = {
            'is_archived': sketch.get_status.status == 'archived',
            'sketch_id': sketch.id,
            'sketch_name': sketch.name,
        }
        schema = {
            'meta': meta,
            'objects': []
        }
        return jsonify(schema)

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Returns:
            A sketch in JSON (instance of flask.wrappers.Response)
        """
        if current_user.admin:
            sketch = Sketch.query.get(sketch_id)
        else:
            sketch = Sketch.query.get_with_acl(sketch_id)

        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        self._sketch = sketch
        form = request.json
        if not form:
            form = request.data

        action = form.get('action', '')
        if action == 'archive':
            if not sketch.has_permission(current_user, 'delete'):
                if not current_user.admin:
                    abort(
                        HTTP_STATUS_CODE_FORBIDDEN,
                        'User does not have sufficient access rights to '
                        'delete a sketch.')

            return self._archive_sketch(sketch)

        if action == 'export':
            if not sketch.has_permission(current_user, 'read'):
                abort(
                    HTTP_STATUS_CODE_FORBIDDEN,
                    'User does not have sufficient access rights to '
                    'read from a sketch.')

            return self._export_sketch(sketch)

        if action == 'unarchive':
            if not sketch.has_permission(current_user, 'delete'):
                if not current_user.admin:
                    abort(
                        HTTP_STATUS_CODE_FORBIDDEN,
                        'User does not have sufficient access rights to '
                        'unarchive a sketch.')

            return self._unarchive_sketch(sketch)

        return abort(
            HTTP_STATUS_CODE_NOT_FOUND,
            'The action: [{0:s}] is not supported.'.format(action))

    def _get_all_events_with_a_label(self, label, sketch):
        """Returns a DataFrame with events in a sketch with a certain label.

        Args:
            label (string): the label string to search for.
            sketch (timesketch.models.sketch.Sketch): a sketch object.

        Returns:
            pd.DataFrame: a pandas DataFrame with all the events in the
                datastore that have a label.
        """
        query_dsl = {
            'query': {
                'nested': {
                    'path': 'timesketch_label',
                    'query': {
                        'bool': {
                            'must': [{
                                'term': {
                                    'timesketch_label.name': label
                                }
                            }, {
                                'term': {
                                    'timesketch_label.sketch_id': sketch.id
                                }
                            }]
                        }
                    }
                }
            }
        }
        result = self.datastore.search(
            sketch_id=sketch.id,
            query_string='',
            query_filter=self.DEFAULT_QUERY_FILTER,
            query_dsl=json.dumps(query_dsl),
            indices=self.sketch_indices)

        return export.query_results_to_dataframe(result, sketch)

    def _export_events_with_comments(self, sketch, zip_file):
        """Export all events that have comments and store in a ZIP file."""
        db_events = Event.query.filter_by(sketch_id=sketch.id).all()
        lines = []
        for db_event in db_events:
            for comment in db_event.comments:
                line = {
                    '_index': db_event.searchindex.index_name,
                    '_id': db_event.document_id,
                    'comment': comment.comment,
                    'comment_date': comment.created_at,
                }
                if not comment.user:
                    line['username'] = 'System'
                else:
                    line['username'] = comment.user.username
                lines.append(line)
        db_frame = pd.DataFrame(lines)

        size = db_frame.shape[0]
        if not size:
            return

        event_frame = self._get_all_events_with_a_label('__ts_comment', sketch)
        frame = event_frame.merge(db_frame, on=['_index', '_id'])

        string_io = io.StringIO()
        frame.to_csv(string_io, index=False)
        string_io.seek(0)
        zip_file.writestr(
            'events/events_with_comments.csv', data=string_io.read())

    def _export_starred_events(self, sketch, zip_file):
        """Export all events that have been starred and store in a ZIP file."""
        event_frame = self._get_all_events_with_a_label('__ts_star', sketch)

        string_io = io.StringIO()
        event_frame.to_csv(string_io, index=False)
        string_io.seek(0)
        zip_file.writestr(
            'events/starred_events.csv', data=string_io.read())

    def _export_tagged_events(self, sketch, zip_file):
        """Export all events that have been tagged and store in a ZIP file."""
        result = self.datastore.search(
            sketch_id=sketch.id,
            query_string='_exists_:tag',
            query_filter=self.DEFAULT_QUERY_FILTER,
            query_dsl='',
            indices=self.sketch_indices)

        fh = export.query_results_to_filehandle(result, sketch)
        zip_file.writestr(
            'events/tagged_events.csv', data=fh.read())

        parameters = {
            'limit': 100,
            'field': 'tag',
        }
        result_obj, meta = utils.run_aggregator(
            sketch.id, aggregator_name='field_bucket',
            aggregator_parameters=parameters)

        zip_file.writestr(
            'events/tagged_event_stats.meta', data=json.dumps(meta))

        html = result_obj.to_chart(
            chart_name='hbarchart',
            chart_title='Top 100 identified tags',
            interactive=True, as_html=True)
        zip_file.writestr(
            'events/tagged_event_stats.html', data=html)

        string_io = io.StringIO()
        data_frame = result_obj.to_pandas()
        data_frame.to_csv(string_io, index=False)
        string_io.seek(0)
        zip_file.writestr(
            'events/tagged_event_stats.csv', data=string_io.read())

    def _export_sketch(self, sketch):
        """Returns a ZIP file with the exported content of a sketch."""
        file_object = io.BytesIO()
        sketch_is_archived = sketch.get_status.status == 'archived'

        if sketch_is_archived:
            _ = self._unarchive_sketch(sketch)

        story_exporter = story_export_manager.StoryExportManager.get_exporter(
            'html')

        meta = {
            'user': current_user.username,
            'time': datetime.datetime.utcnow().isoformat(),
            'sketch_id': sketch.id,
            'sketch_name': sketch.name,
            'sketch_description': sketch.description,
            'timesketch_version': version.get_version(),
        }

        with zipfile.ZipFile(file_object, mode='w') as zip_file:
            zip_file.writestr('METADATA', data=json.dumps(meta))

            for story in sketch.stories:
                export.export_story(story, sketch, story_exporter, zip_file)

            for aggregation in sketch.aggregations:
                export.export_aggregation(aggregation, sketch, zip_file)

            for view in sketch.views:
                self._export_view(view, sketch, zip_file)

            for group in sketch.aggregationgroups:
                export.export_aggregation_group(group, sketch, zip_file)

            self._export_events_with_comments(sketch, zip_file)
            self._export_starred_events(sketch, zip_file)
            self._export_tagged_events(sketch, zip_file)

            # TODO (kiddi): Add in aggregation group support.

        if sketch_is_archived:
            _ = self._archive_sketch(sketch)

        file_object.seek(0)
        return send_file(
            file_object, mimetype='zip',
            attachment_filename='timesketch_export.zip')

    def _export_view(self, view, sketch, zip_file):
        """Export a view from a sketch and write it to a ZIP file.

        Args:
            view (timesketch.models.sketch.View): a View object.
            sketch (timesketch.models.sketch.Sketch): a sketch object.
            zip_file (ZipFile): a zip file handle that can be used to write
                content to.
        """
        name = '{0:04d}_{1:s}'.format(view.id, view.name)

        query_filter = json.loads(view.query_filter)
        if not query_filter:
            query_filter = self.DEFAULT_QUERY_FILTER

        indices = query_filter.get('indices', self.sketch_indices)
        if not indices or '_all' in indices:
            indices = self.sketch_indices

        # Ignoring the size limits in views to reduce the amount of queries
        # needed to get all the data.
        query_filter['terminate_after'] = 10000
        query_filter['size'] = 10000

        query_dsl = view.query_dsl
        if query_dsl:
            query_dict = json.loads(query_dsl)
            if not query_dict:
                query_dsl = None

        result = self.datastore.search(
            sketch_id=sketch.id,
            query_string=view.query_string,
            query_filter=query_filter,
            query_dsl=query_dsl,
            enable_scroll=True,
            indices=indices)

        scroll_id = result.get('_scroll_id', '')
        if scroll_id:
            data_frame = export.query_results_to_dataframe(result, sketch)

            total_count = result.get(
                'hits', {}).get('total', {}).get('value', 0)

            if isinstance(total_count, str):
                try:
                    total_count = int(total_count, 10)
                except ValueError:
                    total_count = 0

            event_count = len(result['hits']['hits'])

            while event_count < total_count:
                # pylint: disable=unexpected-keyword-arg
                result = self.datastore.client.scroll(
                    scroll_id=scroll_id, scroll='1m')
                event_count += len(result['hits']['hits'])
                add_frame = export.query_results_to_dataframe(result, sketch)
                if add_frame.shape[0]:
                    data_frame = pd.concat([data_frame, add_frame], sort=False)
                else:
                    logger.warning(
                        'Data Frame returned from a search operation was '
                        'empty, count {0:d} out of {1:d} total. Query is: '
                        '"{2:s}"'.format(
                            event_count, total_count,
                            view.query_string or query_dsl))

            fh = io.StringIO()
            data_frame.to_csv(fh, index=False)
            fh.seek(0)
        else:
            fh = export.query_results_to_filehandle(result, sketch)

        zip_file.writestr(
            'views/{0:s}.csv'.format(name), data=fh.read())

        if not view.user:
            username = 'System'
        else:
            username = view.user.username
        meta = {
            'name': view.name,
            'view_id': view.id,
            'description': view.description,
            'query_string': view.query_string,
            'query_filter': view.query_filter,
            'query_dsl': view.query_dsl,
            'username': username,
            'sketch_id': view.sketch_id,
        }
        zip_file.writestr(
            'views/{0:s}.meta'.format(name), data=json.dumps(meta))

    def _unarchive_sketch(self, sketch):
        """Unarchives a sketch by opening up all indices and removing labels.

        Args:
            sketch: Instance of timesketch.models.sketch.Sketch
        """
        if sketch.get_status.status != 'archived':
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to unarchive a sketch that wasn\'t already archived '
                '(sketch: {0:d})'.format(sketch.id))

        sketch.set_status(status='ready')

        indexes_to_open = []
        for timeline in sketch.timelines:
            if timeline.get_status.status != 'archived':
                continue
            timeline.set_status(status='ready')
            search_index = timeline.searchindex
            search_index.set_status(status='ready')
            indexes_to_open.append(search_index.index_name)

        # TODO (kiddi): Move this to lib/datastores/elastic.py.
        if indexes_to_open:
            self.datastore.client.indices.open(','.join(indexes_to_open))
        return HTTP_STATUS_CODE_OK

    def _archive_sketch(self, sketch):
        """Unarchives a sketch by opening up all indices and removing labels.

        Args:
            sketch: Instance of timesketch.models.sketch.Sketch
        """
        if sketch.get_status.status == 'archived':
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to archive a sketch that was already archived '
                '(sketch: {0:d})'.format(sketch.id))

        labels_to_prevent_deletion = current_app.config.get(
            'LABELS_TO_PREVENT_DELETION', [])

        for label in labels_to_prevent_deletion:
            if sketch.has_label(label):
                abort(
                    HTTP_STATUS_CODE_FORBIDDEN,
                    'A sketch with the label {0:s} cannot be '
                    'archived.'.format(label))

        sketch.set_status(status='archived')

        # Go through all timelines in a sketch.
        #    Each timeline has only a single search index, however
        #    each search index can be part of multiple timelines.
        #    Only archive a search index if all of it's timelines
        #    are archived.
        indexes_to_close = []
        for timeline in sketch.timelines:
            if timeline.get_status.status != 'ready':
                continue
            timeline.set_status(status='archived')
            search_index = timeline.searchindex

            if not all([
                    x.get_status.status == 'archived'
                    for x in search_index.timelines]):
                continue
            search_index.set_status(status='archived')
            indexes_to_close.append(search_index.index_name)

        # TODO (kiddi): Move this to lib/datastores/elastic.py.
        if indexes_to_close:
            self.datastore.client.indices.close(','.join(indexes_to_close))
        return HTTP_STATUS_CODE_OK



"""Tests for expert sessionizers."""

from __future__ import unicode_literals

import unittest
import mock

from timesketch.lib.analyzers.expert_sessionizers import \
    WebActivitySessionizerSketchPlugin
from timesketch.lib.analyzers.expert_sessionizers import \
    SSHBruteforceSessionizerSketchPlugin
from timesketch.lib.analyzers.base_sessionizer_test import _create_mock_event
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore

class TestWebActivitySessionizerPlugin(BaseTest):
    """Tests the functionality of the web activity sessionizing sketch
    analyzer."""

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_event_type(self):
        """Test the mocking of events returns an event stream that matches the
        query for the analyzer."""
        index = 'test_index'
        sketch_id = 1
        analyzer = WebActivitySessionizerSketchPlugin(index, sketch_id)
        analyzer.datastore.client = mock.Mock()
        datastore = analyzer.datastore

        _create_mock_event(datastore,
                           0,
                           2,
                           source_attrs={'source_short': 'WEBHIST'})

        message = analyzer.run()
        self.assertEqual(
            message, 'Sessionizing completed, number of session created: 1')

        event1 = datastore.event_store['0']
        self.assertEqual(event1['_source']['source_short'], 'WEBHIST')
        self.assertEqual(event1['_source']['session_id'],
                         {analyzer.session_type: 1})
        event2 = datastore.event_store['101']
        self.assertEqual(event2['_source']['source_short'], 'WEBHIST')
        self.assertEqual(event2['_source']['session_id'],
                         {analyzer.session_type: 1})



class TestSSHBruteforceSessionizerPlugin(BaseTest):
    """Tests the functionality of the SSH bruteforce attack sessionizing sketch
    analyzer."""

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_event_type(self):
        """Test the mocking of events returns an event stream that matches the
        query for the analyzer."""
        index = 'test_index'
        sketch_id = 1
        analyzer = SSHBruteforceSessionizerSketchPlugin(index, sketch_id)
        analyzer.datastore.client = mock.Mock()
        datastore = analyzer.datastore

        _create_mock_event(
            datastore,
            0,
            2,
            source_attrs={'reporter': 'sshd',
                          'message': '[sshd] [0]: Invalid user ' \
                          'NoSuchUser from 0.0.0.0 port 0'})

        message = analyzer.run()
        self.assertEqual(
            message,
            'Sessionizing completed, number of session created: 1')

        test_message = '[sshd] [0]: Invalid user NoSuchUser from 0.0.0.0 ' \
            'port 0'
        # pylint: disable=unexpected-keyword-arg
        event1 = datastore.event_store['0']
        self.assertEqual(event1['_source']['reporter'], 'sshd')
        self.assertEqual(event1['_source']['message'], test_message)
        self.assertEqual(event1['_source']['session_id'],
                         {analyzer.session_type: 1})

        event2 = datastore.event_store['1']
        self.assertEqual(event2['_source']['reporter'], 'sshd')
        self.assertEqual(event2['_source']['message'], test_message)
        self.assertEqual(event2['_source']['session_id'],
                         {analyzer.session_type: 1})

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the user model."""

from __future__ import unicode_literals

import codecs

import six

from flask_bcrypt import generate_password_hash
from flask_bcrypt import check_password_hash
from flask_login import UserMixin
from sqlalchemy.types import Boolean
from sqlalchemy import Column
from sqlalchemy import PrimaryKeyConstraint
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Table
from sqlalchemy import Unicode
from sqlalchemy import UnicodeText
from sqlalchemy.orm import backref
from sqlalchemy.orm import relationship

from timesketch.models import BaseModel
from timesketch.models.annotations import LabelMixin
from timesketch.models.annotations import StatusMixin

# Helper table for Groups many-to-many relationship.
user_group = Table('user_group', BaseModel.metadata,
                   Column('user_id', Integer(), ForeignKey('user.id')),
                   Column('group_id', Integer(), ForeignKey('group.id')),
                   PrimaryKeyConstraint('user_id', 'group_id'))


class User(UserMixin, BaseModel):
    """Implements the User model."""

    username = Column(Unicode(255), unique=True)
    password = Column(Unicode(128))
    name = Column(Unicode(255))
    email = Column(Unicode(255))
    active = Column(Boolean(), default=True)
    admin = Column(Boolean(), default=False)
    sketches = relationship('Sketch', backref='user', lazy='dynamic')
    searchindices = relationship(
        'SearchIndex', backref='user', lazy='dynamic')
    timelines = relationship('Timeline', backref='user', lazy='dynamic')
    views = relationship('View', backref='user', lazy='dynamic')
    stories = relationship('Story', backref='user', lazy='dynamic')
    aggregations = relationship('Aggregation', backref='user', lazy='dynamic')
    aggregationgroups = relationship(
        'AggregationGroup', backref='user', lazy='dynamic')
    my_groups = relationship('Group', backref='user', lazy='dynamic')
    groups = relationship(
        'Group',
        secondary=user_group,
        backref=backref('users', lazy='dynamic'))

    def __init__(self, username, name=None):
        """Initialize the User object.

        Args:
            username: Username for the user
            name: Name of the user
        """
        super(User, self).__init__()
        self.username = username
        self.name = name
        if not name:
            self.name = username

    def set_password(self, plaintext, rounds=12):
        """Sets the password for the user. The password hash is created with the
        Bcrypt python library (http://www.mindrot.org/projects/py-bcrypt/).

        Args:
            plaintext: The plaintext password to hash
            rounds: Number of rounds to use for the bcrypt hashing
        """
        password_hash = generate_password_hash(plaintext, rounds)
        if isinstance(password_hash, six.binary_type):
            password_hash = codecs.decode(password_hash, 'utf-8')
        self.password = password_hash

    def check_password(self, plaintext):
        """Check a plaintext password against a stored password hash.

        Args:
            plaintext: A plaintext password

        Returns:
            A boolean value indicating if the plaintext password matches the
            stored password hash.
        """
        return check_password_hash(self.password, plaintext)


class Group(LabelMixin, StatusMixin, BaseModel):
    """Implements the Group model."""

    name = Column(Unicode(255), unique=True)
    display_name = Column(Unicode(255))
    description = Column(UnicodeText())
    user_id = Column(Integer, ForeignKey('user.id'))

    def __init__(self, name, display_name=None, description=None, user=None):
        """Initialize the Group object.

        Args:
            name: Name of the group
            display_name: User friendly name of the group
            description: Description of the group
            user: Creator (instance of timesketch.models.user.User)
        """
        super(Group, self).__init__()
        self.name = name
        self.display_name = display_name or name
        self.description = description or name
        self.user = user



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains a class for managing story exporters."""

from __future__ import unicode_literals


class StoryExportManager(object):
    """The story export manager."""

    _class_registry = {}

    @classmethod
    def clear_registration(cls):
        """Clears all analyzer registration."""
        cls._class_registry = {}

    @classmethod
    def get_exporters(cls, exporter_types=None):
        """Retrieves the registered story exporters.

        Args:
            exporter_types (list): List of exporter types.

        Yields:
            tuple: containing:
                str: the uniquely identifying name of the story exporter.
                type: the exporter class.
        """
        for exporter_type, exporter_class in cls._class_registry.items():
            if exporter_types and exporter_type not in exporter_types:
                continue
            yield exporter_type, exporter_class

    @classmethod
    def get_formats(cls):
        """Retrieves the registered format types."""
        return cls._class_registry.keys()

    @classmethod
    def get_exporter(cls, exporter_type):
        """Retrieves a class object of a specific exporter.

        Args:
            exporter_type (str): name of the exporter to retrieve.

        Returns:
            StoryExporter class object.
        """
        return cls._class_registry.get(exporter_type.lower())

    @classmethod
    def register_exporter(cls, exporter_class):
        """Registers an exporter class.

        The exporter classes are identified by their lower export format.

        Args:
            exporter_class (type): the exporter class to register.

        Raises:
            KeyError: if class is already set for the corresponding name.
        """
        exporter_type = exporter_class.EXPORT_FORMAT.lower()
        if exporter_type in cls._class_registry:
            raise KeyError('Class already set for name: {0:s}.'.format(
                exporter_class.EXPORT_FORMAT))

        cls._class_registry[exporter_type] = exporter_class



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Event resources for version 1 of the Timesketch API."""

import codecs
import datetime
import hashlib
import json
import logging
import math
import time
import six

import dateutil
from elasticsearch.exceptions import RequestError
import numpy as np
import pandas as pd

from flask import jsonify
from flask import request
from flask import abort
from flask_restful import Resource
from flask_restful import reqparse
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import resources
from timesketch.lib import forms
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK
from timesketch.lib.definitions import HTTP_STATUS_CODE_CREATED
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.models import db_session
from timesketch.models.sketch import Event
from timesketch.models.sketch import SearchIndex
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import Timeline


logger = logging.getLogger('timesketch.event_api')


def _tag_event(row, tag_dict, tags_to_add, datastore, flush_interval):
    """Tag each event from a dataframe with tags.

    Args:
        row (np.Series): a single row of data with existing tags and
            information about the event in order to be able to add
            tags to it.
        tag_dict (dict): a dict that contains information to be returned
            by the API call to the user.
        tags_to_add (list[str]): a list of strings of tags to add to each
            event.
        datastore (elastic.ElasticsearchDataStore): the datastore object.
        flush_interval (int): the number of events to import before a bulk
            update is done with the datastore.
    """
    tag_dict['events_processed_by_api'] += 1
    existing_tags = set()

    if 'tag' in row:
        tag = row['tag']
        if isinstance(tag, (list, tuple)):
            existing_tags = set(tag)

        new_tags = list(set().union(existing_tags, set(tags_to_add)))
    else:
        new_tags = tags_to_add

    if set(existing_tags) == set(new_tags):
        return

    datastore.import_event(
        index_name=row['_index'], event_type=row['_type'],
        event_id=row['_id'], event={'tag': new_tags},
        flush_interval=flush_interval)

    tag_dict['tags_applied'] += len(new_tags)
    tag_dict['number_of_events_with_added_tags'] += 1


class EventCreateResource(resources.ResourceMixin, Resource):
    """Resource to create an annotation for an event."""

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.
        Handler for /api/v1/sketches/:sketch_id/event/create/

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            An annotation in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'write'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have write access controls on sketch.')

        form = request.json
        if not form:
            form = request.data

        timeline_name = 'sketch specific timeline'
        index_name_seed = 'timesketch_{0:d}'.format(sketch_id)
        event_type = 'user_created_event'

        date_string = form.get('date_string')
        if not date_string:
            date = datetime.datetime.utcnow().isoformat()
        else:
            # derive datetime from timestamp:
            try:
                date = dateutil.parser.parse(date_string)
            except (dateutil.parser.ParserError, OverflowError) as e:
                logger.error('Unable to convert date string', exc_info=True)
                abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'Unable to add event, not able to convert the date '
                    'string. Was it properly formatted? Error: '
                    '{0!s}'.format(e))

        timestamp = int(
            time.mktime(date.utctimetuple())) * 1000000
        timestamp += date.microsecond

        event = {
            'datetime': date_string,
            'timestamp': timestamp,
            'timestamp_desc': form.get('timestamp_desc', 'Event Happened'),
            'message': form.get('message', 'No message string'),
        }

        attributes = form.get('attributes', {})
        if not isinstance(attributes, dict):
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to add an event where the attributes are not a '
                'dict object.')

        event.update(attributes)

        tag = form.get('tag', [])
        if not isinstance(tag, list):
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to add an event where the tags are not a '
                'list of strings.')

        if tag and any([not isinstance(x, str) for x in tag]):
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to add an event where the tags are not a '
                'list of strings.')

        event['tag'] = tag

        # We do not need a human readable filename or
        # datastore index name, so we use UUIDs here.
        index_name = hashlib.md5(index_name_seed.encode()).hexdigest()
        if six.PY2:
            index_name = codecs.decode(index_name, 'utf-8')

        # Try to create index
        try:
            # Create the index in Elasticsearch (unless it already exists)
            self.datastore.create_index(
                index_name=index_name,
                doc_type=event_type)

            # Create the search index in the Timesketch database
            searchindex = SearchIndex.get_or_create(
                name=timeline_name,
                description='internal timeline for user-created events',
                user=current_user,
                index_name=index_name)
            searchindex.grant_permission(
                permission='read', user=current_user)
            searchindex.grant_permission(
                permission='write', user=current_user)
            searchindex.grant_permission(
                permission='delete', user=current_user)
            searchindex.set_status('ready')
            db_session.add(searchindex)
            db_session.commit()

            timeline = None
            if sketch and sketch.has_permission(current_user, 'write'):
                self.datastore.import_event(
                    index_name,
                    event_type,
                    event,
                    flush_interval=1)

                timeline = Timeline.get_or_create(
                    name=searchindex.name,
                    description=searchindex.description,
                    sketch=sketch,
                    user=current_user,
                    searchindex=searchindex)

                if timeline not in sketch.timelines:
                    sketch.timelines.append(timeline)

                timeline.set_status('ready')
                db_session.add(timeline)
                db_session.commit()

            # Return Timeline if it was created.
            # pylint: disable=no-else-return
            if timeline:
                return self.to_json(
                    timeline, status_code=HTTP_STATUS_CODE_CREATED)
            else:
                return self.to_json(
                    searchindex, status_code=HTTP_STATUS_CODE_CREATED)

        # TODO: Can this be narrowed down, both in terms of the scope it
        # applies to, as well as not to catch a generic exception.
        except Exception as e:  # pylint: disable=broad-except
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Failed to add event ({0!s})'.format(e))


class EventResource(resources.ResourceMixin, Resource):
    """Resource to get a single event from the datastore.

    HTTP Args:
        searchindex_id: The datastore searchindex id as string
        event_id: The datastore event id as string
    """
    def __init__(self):
        super(EventResource, self).__init__()
        self.parser = reqparse.RequestParser()
        self.parser.add_argument(
            'searchindex_id', type=six.text_type, required=True)
        self.parser.add_argument('event_id', type=six.text_type, required=True)

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.
        Handler for /api/v1/sketches/:sketch_id/event/

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            JSON of the datastore event
        """

        args = self.parser.parse_args()
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')

        searchindex_id = args.get('searchindex_id')
        searchindex = SearchIndex.query.filter_by(
            index_name=searchindex_id).first()
        if not searchindex:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Search index not found for this event.')
        if searchindex.get_status.status == 'deleted':
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to query event on a closed search index.')

        event_id = args.get('event_id')
        indices = [
            t.searchindex.index_name for t in sketch.timelines
            if t.get_status.status.lower() == 'ready']

        # Check if the requested searchindex is part of the sketch
        if searchindex_id not in indices:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Search index ID ({0!s}) does not belong to the list '
                'of indices'.format(searchindex_id))

        result = self.datastore.get_event(searchindex_id, event_id)

        event = Event.query.filter_by(
            sketch=sketch, searchindex=searchindex,
            document_id=event_id).first()

        # Comments for this event
        comments = []
        if event:
            for comment in event.comments:
                if not comment.user:
                    username = 'System'
                else:
                    username = comment.user.username
                comment_dict = {
                    'user': {
                        'username': username,
                    },
                    'created_at': comment.created_at,
                    'comment': comment.comment
                }
                comments.append(comment_dict)

        schema = {
            'meta': {
                'comments': comments
            },
            'objects': result['_source']
        }
        return jsonify(schema)


class EventTaggingResource(resources.ResourceMixin, Resource):
    """Resource to fetch and set tags to an event."""

    # The number of events to bulk together for each query.
    EVENT_CHUNK_SIZE = 1000

    # The maximum number of events to tag in a single request.
    MAX_EVENTS_TO_TAG = 100000

    # The size of the buffer before a bulk update in ES takes place.
    BUFFER_SIZE_FOR_ES_BULK_UPDATES = 10000

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            An annotation in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            msg = 'No sketch found with this ID.'
            abort(HTTP_STATUS_CODE_NOT_FOUND, msg)

        if not sketch.has_permission(current_user, 'write'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN, (
                    'User does not have sufficient access rights to '
                    'modify a sketch.'))

        form = request.json
        if not form:
            form = request.data

        tag_dict = {
            'events_processed_by_api': 0,
            'number_of_events_with_added_tags': 0,
            'tags_applied': 0,
        }
        datastore = self.datastore

        try:
            tags_to_add = json.loads(form.get('tag_string', ''))
        except json.JSONDecodeError as e:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to read the tags, with error: {0!s}'.format(e))

        if not isinstance(tags_to_add, list):
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST, 'Tags need to be a list')

        if not all([isinstance(x, str) for x in tags_to_add]):
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Tags need to be a list of strings')

        events = form.get('events', [])
        event_df = pd.DataFrame(events)

        for field in ['_id', '_type', '_index']:
            if field not in event_df:
                abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'Events need to have a [{0:s}] field associated '
                    'to it.'.format(field))
            if any(event_df[field].isna()):
                abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'All events need to have a [{0:s}] field '
                    'set, it cannot have a non-value.'.format(field))

        # Remove any potential extra fields from the events.
        event_df = event_df[['_id', '_type', '_index']]

        tag_df = pd.DataFrame()

        event_size = event_df.shape[0]
        if event_size > self.MAX_EVENTS_TO_TAG:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Cannot tag more than {0:d} events in a single '
                'request'.format(self.MAX_EVENTS_TO_TAG))

        tag_dict['number_of_events_passed_to_api'] = event_size

        errors = []

        verbose = form.get('verbose', False)
        if verbose:
            tag_dict['number_of_indices'] = len(event_df['_index'].unique())
            time_tag_gathering_start = time.time()

        for _index in event_df['_index'].unique():
            index_slice = event_df[event_df['_index'] == _index]
            index_size = index_slice.shape[0]

            if verbose:
                tag_dict.setdefault('index_count', {})
                tag_dict['index_count'][_index] = index_size

            if index_size <= self.EVENT_CHUNK_SIZE:
                chunks = 1
            else:
                chunks = math.ceil(index_size / self.EVENT_CHUNK_SIZE)

            tags = []
            for index_chunk in np.array_split(
                    index_slice['_id'].unique(), chunks):
                should_list = [{'match': {'_id': x}} for x in index_chunk]
                query_body = {
                    'query': {
                        'bool': {
                            'should': should_list
                        }
                    }
                }

                # Adding a small buffer to make sure all results are captured.
                size = len(should_list) + 100
                query_body['size'] = size
                query_body['terminate_after'] = size

                try:
                    # pylint: disable=unexpected-keyword-arg
                    if datastore.version.startswith('6'):
                        search = datastore.client.search(
                            body=json.dumps(query_body),
                            index=[_index],
                            _source_include=['tag'],
                            search_type='query_then_fetch'
                        )
                    else:
                        search = datastore.client.search(
                            body=json.dumps(query_body),
                            index=[_index],
                            _source_includes=['tag'],
                            search_type='query_then_fetch'
                        )

                except RequestError as e:
                    logger.error('Unable to query for events', exc_info=True)
                    errors.append(
                        'Unable to query for events, {0!s}'.format(e))
                    abort(
                        HTTP_STATUS_CODE_BAD_REQUEST,
                        'Unable to query events, {0!s}'.format(e))

                for result in search['hits']['hits']:
                    tag = result.get('_source', {}).get('tag', [])
                    if not tag:
                        continue
                    tags.append({'_id': result.get('_id'), 'tag': tag})

            if not tags:
                continue
            tag_df = pd.concat([tag_df, pd.DataFrame(tags)])

        if tag_df.shape[0]:
            event_df = event_df.merge(tag_df, on='_id', how='left')

        if verbose:
            tag_dict[
                'time_to_gather_tags'] = time.time() - time_tag_gathering_start
            tag_dict['number_of_events'] = len(events)

            if tag_df.shape[0]:
                tag_dict['number_of_events_in_tag_frame'] = tag_df.shape[0]

            if 'tag' in event_df:
                current_tag_events = event_df[~event_df['tag'].isna()].shape[0]
                tag_dict['number_of_events_with_tags'] = current_tag_events
            else:
                tag_dict['number_of_events_with_tags'] = 0

            tag_dict['tags_to_add'] = tags_to_add
            time_tag_start = time.time()

        if event_size > datastore.DEFAULT_FLUSH_INTERVAL:
            flush_interval = self.BUFFER_SIZE_FOR_ES_BULK_UPDATES
        else:
            flush_interval = datastore.DEFAULT_FLUSH_INTERVAL
        _ = event_df.apply(
            _tag_event, axis=1, tag_dict=tag_dict, tags_to_add=tags_to_add,
            datastore=datastore, flush_interval=flush_interval)
        datastore.flush_queued_events()

        if verbose:
            tag_dict['time_to_tag'] = time.time() - time_tag_start

        if errors:
            tag_dict['errors'] = errors

        schema = {
            'meta': tag_dict,
            'objects': []}
        response = jsonify(schema)
        response.status_code = HTTP_STATUS_CODE_OK
        return response


class EventAnnotationResource(resources.ResourceMixin, Resource):
    """Resource to create an annotation for an event."""

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            An annotation in JSON (instance of flask.wrappers.Response)
        """
        form = forms.EventAnnotationForm.build(request)
        if not form.validate_on_submit():
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST, 'Unable to validate form data.')

        annotations = []
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'write'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have write access controls on sketch.')

        indices = [
            t.searchindex.index_name for t in sketch.timelines
            if t.get_status.status.lower() == 'ready']
        annotation_type = form.annotation_type.data
        events = form.events.raw_data

        for _event in events:
            searchindex_id = _event['_index']
            searchindex = SearchIndex.query.filter_by(
                index_name=searchindex_id).first()
            event_id = _event['_id']
            event_type = _event['_type']

            if searchindex_id not in indices:
                abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'Search index ID ({0!s}) does not belong to the list '
                    'of indices'.format(searchindex_id))

            # Get or create an event in the SQL database to have something
            # to attach the annotation to.
            event = Event.get_or_create(
                sketch=sketch,
                searchindex=searchindex,
                document_id=event_id)

            # Add the annotation to the event object.
            if 'comment' in annotation_type:
                annotation = Event.Comment(
                    comment=form.annotation.data, user=current_user)
                event.comments.append(annotation)
                self.datastore.set_label(
                    searchindex_id,
                    event_id,
                    event_type,
                    sketch.id,
                    current_user.id,
                    '__ts_comment',
                    toggle=False)

            elif 'label' in annotation_type:
                annotation = Event.Label.get_or_create(
                    label=form.annotation.data, user=current_user)
                if annotation not in event.labels:
                    event.labels.append(annotation)
                toggle = False
                if '__ts_star' or '__ts_hidden' in form.annotation.data:
                    toggle = True
                if form.remove.data:
                    toggle = True
                self.datastore.set_label(
                    searchindex_id,
                    event_id,
                    event_type,
                    sketch.id,
                    current_user.id,
                    form.annotation.data,
                    toggle=toggle)

            else:
                abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'Annotation type needs to be either label or comment, '
                    'not {0!s}'.format(annotation_type))

            annotations.append(annotation)
            # Save the event to the database
            db_session.add(event)
            db_session.commit()

        return self.to_json(
            annotations, status_code=HTTP_STATUS_CODE_CREATED)


class CountEventsResource(resources.ResourceMixin, Resource):
    """Resource to number of events for sketch timelines."""

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            Number of events in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')
        indices = [
            t.searchindex.index_name for t in sketch.active_timelines
            if t.get_status.status != 'archived'
        ]
        count = self.datastore.count(indices)
        meta = dict(count=count)
        schema = dict(meta=meta, objects=[])
        return jsonify(schema)



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for utils."""

from __future__ import unicode_literals

import re

from timesketch.lib.testlib import BaseTest
from timesketch.lib.utils import get_validated_indices
from timesketch.lib.utils import random_color


class TestUtils(BaseTest):
    """Tests for the functionality on the utils module."""

    def test_random_color(self):
        """Test to generate a random color."""
        color = random_color()
        self.assertTrue(re.match('^[0-9a-fA-F]{6}$', color))

    def test_get_validated_indices(self):
        """Test for validating indices."""
        sketch = self.sketch1
        sketch_indices = [t.searchindex.index_name for t in sketch.timelines]
        valid_indices = ['test']
        invalid_indices = ['test', 'fail']
        self.assertListEqual(sketch_indices,
                             get_validated_indices(valid_indices,
                                                   sketch_indices))
        self.assertFalse('fail' in get_validated_indices(
            invalid_indices, sketch_indices))



# Copyright 2018 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for analysis manager."""

from __future__ import unicode_literals

from timesketch.lib.testlib import BaseTest
from timesketch.lib.analyzers import manager


class MockAnalyzer(object):
    """Mock analyzer class,"""
    NAME = 'MockAnalyzer'

    DEPENDENCIES = frozenset()


class MockAnalyzer2(object):
    """Mock analyzer class,"""
    NAME = 'MockAnalyzer2'

    DEPENDENCIES = frozenset(['MockAnalyzer'])


class MockAnalyzer3(object):
    """Mock analyzer class,"""
    NAME = 'MockAnalyzer3'

    DEPENDENCIES = frozenset()


class MockAnalyzer4(object):
    """Mock analyzer class,"""
    NAME = 'MockAnalyzer4'

    DEPENDENCIES = frozenset(['MockAnalyzer2', 'MockAnalyzer3'])


class MockAnalyzerFail1(object):
    """Mock analyzer class,"""
    NAME = 'MockAnalyzerFail1'

    DEPENDENCIES = frozenset(['MockAnalyzerFail2'])


class MockAnalyzerFail2(object):
    """Mock analyzer class,"""
    NAME = 'MockAnalyzerFail2'

    DEPENDENCIES = frozenset(['MockAnalyzerFail1'])


class TestAnalysisManager(BaseTest):
    """Tests for the functionality of the manager module."""

    def setUp(self):
        """Set up the tests."""
        super(TestAnalysisManager, self).setUp()
        manager.AnalysisManager.clear_registration()
        manager.AnalysisManager.register_analyzer(MockAnalyzer)

    def test_get_analyzers(self):
        """Test to get analyzer class objects."""
        analyzers = manager.AnalysisManager.get_analyzers()
        analyzer_list = list(analyzers)

        analyzer_dict = {}
        for name, analyzer_class in analyzer_list:
            analyzer_dict[name] = analyzer_class
        self.assertIn('mockanalyzer', analyzer_dict)
        analyzer_class = analyzer_dict.get('mockanalyzer')
        self.assertEqual(analyzer_class, MockAnalyzer)

        manager.AnalysisManager.clear_registration()
        analyzers = manager.AnalysisManager.get_analyzers()
        analyzer_list = list(analyzers)
        self.assertEqual(analyzer_list, [])

        manager.AnalysisManager.register_analyzer(MockAnalyzer)
        manager.AnalysisManager.register_analyzer(MockAnalyzer2)
        manager.AnalysisManager.register_analyzer(MockAnalyzer3)
        manager.AnalysisManager.register_analyzer(MockAnalyzer4)

        analyzers = manager.AnalysisManager.get_analyzers()
        analyzer_names_list = [x for x, _ in analyzers]
        self.assertEqual(len(analyzer_names_list), 4)
        self.assertIn('mockanalyzer', analyzer_names_list)

        # pylint: disable=protected-access
        analyzers_to_run = [
            'mockanalyzer',
            'mockanalyzer2',
            'mockanalyzer3',
            'mockanalyzer4'
        ]
        dependency_tree = manager.AnalysisManager._build_dependencies(
            analyzers_to_run)
        self.assertEqual(len(dependency_tree), 3)
        self.assertIn('mockanalyzer', dependency_tree[0])
        self.assertIn('mockanalyzer3', dependency_tree[0])
        self.assertIn('mockanalyzer2', dependency_tree[1])
        self.assertIn('mockanalyzer4', dependency_tree[2])

        manager.AnalysisManager.clear_registration()
        manager.AnalysisManager.register_analyzer(MockAnalyzerFail1)
        manager.AnalysisManager.register_analyzer(MockAnalyzerFail2)
        with self.assertRaises(KeyError):
            analyzers = manager.AnalysisManager.get_analyzers()
            _ = list(analyzers)

    def test_get_analyzer(self):
        """Test to get analyzer class from registry."""
        analyzer_class = manager.AnalysisManager.get_analyzer('mockanalyzer')
        self.assertEqual(analyzer_class, MockAnalyzer)

    def test_register_analyzer(self):
        """Test so we raise KeyError when analyzer is already registered."""
        self.assertRaises(
            KeyError, manager.AnalysisManager.register_analyzer, MockAnalyzer)



"""Index analyzer plugin for sigma."""
from __future__ import unicode_literals

import logging
import os
import time
import codecs
import elasticsearch

from sigma.backends import elasticsearch as sigma_elasticsearch
import sigma.configuration as sigma_configuration
from sigma.parser import collection as sigma_collection
from timesketch.lib.analyzers import utils

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager


logger = logging.getLogger('timesketch.analyzers.sigma_tagger')


class SigmaPlugin(interface.BaseSketchAnalyzer):
    """Index analyzer for Sigma."""

    NAME = 'sigma'

    _CONFIG_FILE = 'sigma_config.yaml'

    # Path to the directory containing the Sigma Rules to run, relative to
    # this file.
    _RULES_PATH = ''

    def __init__(self, index_name, sketch_id):
        """Initialize the Index Analyzer.

        Args:
            index_name: Elasticsearch index name.
            sketch_id: Sketch ID.
        """
        super(SigmaPlugin, self).__init__(index_name, sketch_id)
        sigma_config_path = interface.get_config_path(self._CONFIG_FILE)
        logger.debug('[sigma] Loading config from {0!s}'.format(
            sigma_config_path))
        with open(sigma_config_path, 'r') as sigma_config_file:
            sigma_config = sigma_config_file.read()
        self.sigma_config = sigma_configuration.SigmaConfiguration(sigma_config)

    def run_sigma_rule(self, query, tag_name):
        """Runs a sigma rule and applies the appropriate tags.

        Args:
            query: elastic search query for events to tag.
            tag_name: tag to apply to matching events.

        Returns:
            int: number of events tagged.
        """
        return_fields = []
        tagged_events_counter = 0
        events = self.event_stream(
            query_string=query, return_fields=return_fields)
        for event in events:
            event.add_tags(['sigma_{0:s}'.format(tag_name)])
            event.commit()
            tagged_events_counter += 1
        return tagged_events_counter

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result.
        """
        sigma_backend = sigma_elasticsearch.ElasticsearchQuerystringBackend(
            self.sigma_config, {})
        tags_applied = {}

        sigma_rule_counter = 0

        rules_path = os.path.join(os.path.dirname(__file__), self._RULES_PATH)


        for dirpath, dirnames, files in os.walk(rules_path):

            if 'deprecated' in dirnames:
                dirnames.remove('deprecated')

            for rule_filename in files:
                if rule_filename.lower().endswith('yml'):

                    # if a sub dir is found, append it to be scanned for rules
                    if os.path.isdir(os.path.join(rules_path, rule_filename)):
                        logger.error(
                            'this is a directory, skipping: {0:s}'.format(
                                rule_filename))
                        continue

                    tag_name, _ = rule_filename.rsplit('.')
                    tags_applied[tag_name] = 0
                    rule_file_path = os.path.join(dirpath, rule_filename)
                    rule_file_path = os.path.abspath(rule_file_path)
                    logger.info('[sigma] Reading rules from {0!s}'.format(
                        rule_file_path))
                    with codecs.open(rule_file_path, 'r', encoding='utf-8',
                                     errors='replace') as rule_file:
                        try:
                            rule_file_content = rule_file.read()
                            parser = sigma_collection.SigmaCollectionParser(
                                rule_file_content, self.sigma_config, None)
                            parsed_sigma_rules = parser.generate(sigma_backend)
                        except NotImplementedError as exception:
                            logger.error(
                                'Error generating rule in file {0:s}: {1!s}'
                                .format(rule_file_path, exception))
                            continue

                        for sigma_rule in parsed_sigma_rules:
                            try:
                                sigma_rule_counter += 1
                                logger.info(
                                    '[sigma] Generated query {0:s}'
                                    .format(sigma_rule))
                                tagged_events_counter = self.run_sigma_rule(
                                    sigma_rule, tag_name)
                                tags_applied[tag_name] += tagged_events_counter
                            except elasticsearch.TransportError as e:
                                logger.error(
                                    'Timeout generating rule in file {0:s}: '
                                    '{1!s} waiting for 10 seconds'.format(
                                        rule_file_path, e), exc_info=True)
                                time.sleep(10) # waiting 10 seconds

        total_tagged_events = sum(tags_applied.values())
        output_string = 'Applied {0:d} tags\n'.format(total_tagged_events)
        for tag_name, tagged_events_counter in tags_applied.items():
            output_string += '* {0:s}: {1:d}\n'.format(
                tag_name, tagged_events_counter)

        if sigma_rule_counter > 0:
            view = self.sketch.add_view(
                view_name='Sigma Rule matches', analyzer_name=self.NAME,
                query_string='tag:"sigma*"')
            agg_params = {
                'field': 'tag',
                'limit': 20,
            }
            agg_obj = self.sketch.add_aggregation(
                name='Top 20 Sigma tags', agg_name='field_bucket',
                agg_params=agg_params, view_id=view.id, chart_type='hbarchart',
                description='Created by the Sigma analyzer')


            story = self.sketch.add_story("Sigma Rule hits")
            story.add_text(
                utils.SIGMA_STORY_HEADER, skip_if_exists=True)

            story.add_text(
                '## Sigma Analyzer.\n\nThe Sigma '
                'analyzer takes Events and matches them with Sigma rules.'
                'In this timeline the analyzer discovered {0:d} '
                'Sigma tags.\n\nThis is a summary of '
                'it\'s findings.'.format(sigma_rule_counter))
            story.add_text(
                'The top 20 most commonly discovered tags were:')
            story.add_aggregation(agg_obj)
            story.add_text(
                'And an overview of all the discovered search terms:')
            story.add_view(view)


        return output_string

class RulesSigmaPlugin(SigmaPlugin):
    """Sigma plugin to run rules."""

    _RULES_PATH = '../../../data/sigma/rules/'

    NAME = 'sigma'

manager.AnalysisManager.register_analyzer(RulesSigmaPlugin)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Attribute resources for version 1 of the Timesketch API."""

import logging

from flask import request
from flask import abort
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import resources
from timesketch.api.v1 import utils
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.models import db_session
from timesketch.models.sketch import Attribute
from timesketch.models.sketch import AttributeValue
from timesketch.models.sketch import Sketch


logger = logging.getLogger('timesketch.sketch_api')


class AttributeResource(resources.ResourceMixin, Resource):
    """Resource for sketch attributes."""

    def _validate_form_entry(self, form, key_to_check):
        """Check a form value and return an error string if applicable.

        Args:
            form (dict): the dict that keeps the form data.
            key_to_check (str): the key in the form data to check against.

        Returns:
            An empty string if there are no issues, otherwise an
            error string to use to abort.
        """
        value = form.get(key_to_check)
        if not value:
            return 'Unable to save an attribute without a {0:s}.'.format(
                key_to_check)

        if not isinstance(value, str):
            return 'Unable to save an attribute without a {0:s}.'.format(
                key_to_check)

        return ''

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.

        Returns:
            An analysis in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'read'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'User does not have read access to sketch')

        return self.to_json(utils.get_sketch_attributes(sketch))

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Returns:
            A string with the response from running the analyzer.
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'write'):
            return abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'User does not have write permission on the sketch.')

        form = request.json
        if not form:
            form = request.data

        if not form:
            return abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'Unable to add or remove an attribute from a '
                'sketch without any data submitted.')

        for check in ['name', 'ontology']:
            error_message = self._validate_form_entry(form, check)
            if error_message:
                return abort(
                    HTTP_STATUS_CODE_BAD_REQUEST, error_message)

        action = form.get('action', '')
        if action not in ('post', 'delete'):
            return abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Action needs to be either "post" or "delete"')

        name = form.get('name')
        ontology = form.get('ontology', 'text')

        if action == 'post':
            values = form.get('values')
            if not values:
                return abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'Missing values from the request.')

            if not isinstance(values, (list, tuple)):
                return abort(
                    HTTP_STATUS_CODE_BAD_REQUEST, 'Values needs to be a list.')

            if any([not isinstance(x, str) for x in values]):
                return abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'All values needs to be stored as strings.')

            for attribute in sketch.attributes:
                if attribute.name == name:
                    return abort(
                        HTTP_STATUS_CODE_BAD_REQUEST,
                        'Unable to add the attribute, it already exists.')

            attribute = Attribute(
                user=current_user,
                sketch=sketch,
                name=name,
                ontology=ontology)
            db_session.add(attribute)
            db_session.commit()

            for value in values:
                attribute_value = AttributeValue(
                    user=current_user,
                    attribute=attribute,
                    value=value)
                attribute.values.append(attribute_value)
                db_session.add(attribute_value)
                db_session.commit()

            db_session.add(attribute)
            db_session.commit()

            return HTTP_STATUS_CODE_OK

        if action != 'delete':
            return abort(
                HTTP_STATUS_CODE_BAD_REQUEST, 'Unable to proceed.')

        for attribute in sketch.attributes:
            if attribute.name != name:
                continue
            for value in attribute.values:
                attribute.values.remove(value)
            sketch.attributes.remove(attribute)
            db_session.commit()

            return HTTP_STATUS_CODE_OK

        return abort(
            HTTP_STATUS_CODE_BAD_REQUEST,
            'Unable to delete the attribute, couldn\'t find it.')



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""View resources for version 1 of the Timesketch API."""

import datetime
import io
import json
import zipfile

from flask import abort
from flask import jsonify
from flask import request
from flask import send_file
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import export
from timesketch.api.v1 import resources
from timesketch.lib import forms
from timesketch.lib.utils import get_validated_indices
from timesketch.lib.definitions import DEFAULT_SOURCE_FIELDS
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.models import db_session
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import View


class ExploreResource(resources.ResourceMixin, Resource):
    """Resource to search the datastore based on a query and a filter."""

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.
        Handler for /api/v1/sketches/:sketch_id/explore/

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            JSON with list of matched events
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')

        if sketch.get_status.status == 'archived':
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to query on an archived sketch.')

        form = forms.ExploreForm.build(request)

        if not form.validate_on_submit():
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to explore data, unable to validate form data')

        # TODO: Remove form and use json instead.
        query_dsl = form.dsl.data
        enable_scroll = form.enable_scroll.data
        scroll_id = form.scroll_id.data
        file_name = form.file_name.data

        query_filter = request.json.get('filter', {})

        return_field_string = form.fields.data
        if return_field_string:
            return_fields = [x.strip() for x in return_field_string.split(',')]
        else:
            return_fields = query_filter.get('fields', [])
            return_fields = [field['field'] for field in return_fields]
            return_fields.extend(DEFAULT_SOURCE_FIELDS)

        sketch_indices = {
            t.searchindex.index_name
            for t in sketch.timelines
            if t.get_status.status.lower() == 'ready'
        }
        if not query_filter:
            query_filter = {}

        indices = query_filter.get('indices', sketch_indices)

        # If _all in indices then execute the query on all indices
        if '_all' in indices:
            indices = sketch_indices

        # Make sure that the indices in the filter are part of the sketch.
        # This will also remove any deleted timeline from the search result.
        indices = get_validated_indices(indices, sketch_indices)

        # Make sure we have a query string or star filter
        if not (form.query.data, query_filter.get('star'),
                query_filter.get('events'), query_dsl):
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'The request needs a query string/DSL and or a star filter.')

        # Aggregate hit count per index.
        index_stats_agg = {
            "indices": {
                "terms": {
                    "field": "_index"
                }
            }
        }

        if file_name:
            file_object = io.BytesIO()

            form_data = {
                'created_at': datetime.datetime.utcnow().isoformat(),
                'created_by': current_user.username,
                'sketch': sketch_id,
                'query': form.query.data,
                'query_dsl': query_dsl,
                'query_filter': query_filter,
                'return_fields': return_fields,
            }
            with zipfile.ZipFile(file_object, mode='w') as zip_file:
                zip_file.writestr('METADATA', data=json.dumps(form_data))
                fh = export.query_to_filehandle(
                    query_string=form.query.data,
                    query_dsl=query_dsl,
                    query_filter=query_filter,
                    indices=indices,
                    sketch=sketch,
                    datastore=self.datastore)
                fh.seek(0)
                zip_file.writestr('query_results.csv', fh.read())
            file_object.seek(0)
            return send_file(
                file_object, mimetype='zip',
                attachment_filename=file_name)

        if scroll_id:
            # pylint: disable=unexpected-keyword-arg
            result = self.datastore.client.scroll(
                scroll_id=scroll_id, scroll='1m')
        else:
            try:
                result = self.datastore.search(
                    sketch_id,
                    form.query.data,
                    query_filter,
                    query_dsl,
                    indices,
                    aggregations=index_stats_agg,
                    return_fields=return_fields,
                    enable_scroll=enable_scroll)
            except ValueError as e:
                abort(
                    HTTP_STATUS_CODE_BAD_REQUEST, e)

        # Get number of matching documents per index.
        count_per_index = {}
        try:
            for bucket in result['aggregations']['indices']['buckets']:
                key = bucket.get('key')
                if key:
                    count_per_index[key] = bucket.get('doc_count')
        except KeyError:
            pass

        # Get labels for each event that matches the sketch.
        # Remove all other labels.
        for event in result['hits']['hits']:
            event['selected'] = False
            event['_source']['label'] = []
            try:
                for label in event['_source']['timesketch_label']:
                    if sketch.id != label['sketch_id']:
                        continue
                    event['_source']['label'].append(label['name'])
                del event['_source']['timesketch_label']
            except KeyError:
                pass

        # Update or create user state view. This is used in the UI to let
        # the user get back to the last state in the explore view.
        view = View.get_or_create(
            user=current_user, sketch=sketch, name='')
        view.query_string = form.query.data
        view.query_filter = json.dumps(query_filter, ensure_ascii=False)
        view.query_dsl = json.dumps(query_dsl, ensure_ascii=False)
        db_session.add(view)
        db_session.commit()

        # Add metadata for the query result. This is used by the UI to
        # render the event correctly and to display timing and hit count
        # information.
        tl_colors = {}
        tl_names = {}
        for timeline in sketch.timelines:
            tl_colors[timeline.searchindex.index_name] = timeline.color
            tl_names[timeline.searchindex.index_name] = timeline.name

        meta = {
            'es_time': result['took'],
            'es_total_count': result['hits']['total'],
            'timeline_colors': tl_colors,
            'timeline_names': tl_names,
            'count_per_index': count_per_index,
            'scroll_id': result.get('_scroll_id', ''),
        }

        # Elasticsearch version 7.x returns total hits as a dictionary.
        # TODO: Refactor when version 6.x has been deprecated.
        if isinstance(meta['es_total_count'], dict):
            meta['es_total_count'] = meta['es_total_count'].get('value', 0)

        schema = {'meta': meta, 'objects': result['hits']['hits']}
        return jsonify(schema)


class QueryResource(resources.ResourceMixin, Resource):
    """Resource to get a query."""

    @login_required
    def post(self, sketch_id):
        """Handles GET request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            A story in JSON (instance of flask.wrappers.Response)
        """
        form = forms.ExploreForm.build(request)
        if not form.validate_on_submit():
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST, 'Unable to validate form data.')
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')
        schema = {
            'objects': [],
            'meta': {}}
        query_string = form.query.data
        query_filter = form.filter.data
        query_dsl = form.dsl.data
        query = self.datastore.build_query(
            sketch.id, query_string, query_filter, query_dsl)
        schema['objects'].append(query)
        return jsonify(schema)



# -*- coding: utf-8 -*-
"""This file contains the plugin for executables in Windows prefetch files."""

from timesketch.lib.analyzers.chain_plugins import interface
from timesketch.lib.analyzers.chain_plugins import manager


class WinPrefetchChainPlugin(interface.BaseChainPlugin):
    """A plugin to chain prefetch records to other events."""

    NAME = 'winprefetch'
    DESCRIPTION = (
        'Plugin to chain prefetch records that show executions on a Windows '
        'system to other events with that executable in it.')

    SEARCH_QUERY = 'data_type:"windows:prefetch:execution"'
    EVENT_FIELDS = ['executable']

    def process_chain(self, base_event):
        """Determine if the extracted event fits the criteria of the plugin.

        Args:
            base_event: an event object (instance of Event).

        Returns:
            boolean to determine whether a chain should be generated from
            the event or not. By default this returns True.
        """
        target = base_event.source.get('executable', '')
        return target.lower().endswith('.exe')

    def get_chained_events(self, base_event):
        """Yields an event that is chained or linked to the base event.

        Args:
            base_event: the base event of the chain, used to construct further
                queries (instance of Event).

        Yields:
            An event (instance of Event) object that is linked or chained to
            the base event, according to the plugin.
        """
        target = base_event.source.get('executable', '')
        if not target:
            return
            yield  # pylint: disable=W0101

        search_query = 'url:"*{0:s}*"'.format(target)
        return_fields = ['url']

        events = self.analyzer_object.event_stream(
            search_query, return_fields=return_fields, scroll=False)
        for event in events:
            url = event.source.get('url', '')
            if target.lower() in url.lower():
                yield event

        lnk_query = 'parser:lnk'
        return_fields = ['link_target']

        events = self.analyzer_object.event_stream(
            lnk_query, return_fields=return_fields, scroll=False)
        for event in events:
            link_target = event.source.get('link_target', '')
            if target.lower() in link_target.lower():
                yield event


manager.ChainPluginsManager.register_plugin(WinPrefetchChainPlugin)



# -*- coding: utf-8 -*-
"""The chain plugins manager object."""

class ChainPluginsManager(object):
    """Class that implements the chain plugins manager."""

    _plugin_classes = {}

    @classmethod
    def deregister_plugin(cls, plugin_class):
        """Deregisters a plugin class.

        The plugin classes are identified based on their lower case name.

        Args:
          plugin_class (type): the class object of the plugin.

        Raises:
          KeyError: if plugin class is not set for the corresponding name.
        """
        plugin_name = plugin_class.NAME.lower()
        if plugin_name not in cls._plugin_classes:
            raise KeyError('Plugin class not set for name: {0:s}.'.format(
                plugin_class.NAME))

        del cls._plugin_classes[plugin_name]

    @classmethod
    def get_plugins(cls, analyzer_object):
        """Retrieves the chain plugins.

        Args:
            analyzer_object: an instance of Sketch object.

        Returns:
            list[type]: list of all chain plugin objects.
        """
        return [plugin_class(analyzer_object) for plugin_class in iter(
            cls._plugin_classes.values())]

    @classmethod
    def register_plugin(cls, plugin_class):
        """Registers a plugin class.

        The plugin classes are identified based on their lower case name.

        Args:
            plugin_class (type): the class object of the plugin.

        Raises:
            KeyError: if plugin class is already set for the corresponding
            name.
        """
        plugin_name = plugin_class.NAME.lower()
        if plugin_name in cls._plugin_classes:
            raise KeyError(('Plugin class already set for : {0:s}.').format(
                plugin_class.NAME))

        cls._plugin_classes[plugin_name] = plugin_class

    @classmethod
    def register_plugins(cls, plugin_classes):
        """Registers plugin classes.

        The plugin classes are identified based on their lower case name.

        Args:
            plugin_classes (list[type]): a list of class objects of the
            plugins.

        Raises:
            KeyError: if plugin class is already set for the corresponding
            name.
        """
        for plugin_class in plugin_classes:
            cls.register_plugin(plugin_class)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sketch resources for version 1 of the Timesketch API."""

import logging

import elasticsearch
import six

from flask import jsonify
from flask import request
from flask import abort
from flask import current_app
from flask_restful import Resource
from flask_restful import reqparse
from flask_login import login_required
from flask_login import current_user
from sqlalchemy import not_

from timesketch.api.v1 import resources
from timesketch.api.v1 import utils
from timesketch.lib import forms
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK
from timesketch.lib.definitions import HTTP_STATUS_CODE_CREATED
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.lib.analyzers import manager as analyzer_manager
from timesketch.lib.aggregators import manager as aggregator_manager
from timesketch.lib.emojis import get_emojis_as_dict
from timesketch.models import db_session
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import SearchTemplate


logger = logging.getLogger('timesketch.sketch_api')


class SketchListResource(resources.ResourceMixin, Resource):
    """Resource for listing sketches."""

    def __init__(self):
        super(SketchListResource, self).__init__()
        self.parser = reqparse.RequestParser()
        self.parser.add_argument(
            'name', type=six.text_type, required=True)
        self.parser.add_argument(
            'description', type=six.text_type, required=False)

    @login_required
    def get(self):
        """Handles GET request to the resource.

        Returns:
            List of sketches (instance of flask.wrappers.Response)
        """
        if current_user.admin:
            sketch_query = Sketch.query
        else:
            sketch_query = Sketch.all_with_acl()

        filtered_sketches = sketch_query.filter(
            not_(Sketch.Status.status == 'deleted'),
            Sketch.Status.parent).order_by(Sketch.updated_at.desc()).all()

        # Just return a subset of the sketch objects to reduce the amount of
        # data returned.
        sketches = []
        for sketch in filtered_sketches:
            sketches.append({
                'name': sketch.name,
                'updated_at': str(sketch.updated_at),
                'user': sketch.user.username,
                'id': sketch.id,
                'status': sketch.get_status.status
            })
        meta = {'current_user': current_user.username}
        return jsonify({'objects': sketches, 'meta': meta})

    @login_required
    def post(self):
        """Handles POST request to the resource.

        Returns:
            A sketch in JSON (instance of flask.wrappers.Response)
        """
        form = forms.NameDescriptionForm.build(request)
        if not form.validate_on_submit():
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST, 'Unable to validate form data.')
        sketch = Sketch(
            name=form.name.data,
            description=form.description.data,
            user=current_user)
        sketch.status.append(sketch.Status(user=None, status='new'))
        db_session.add(sketch)
        db_session.commit()

        # Give the requesting user permissions on the new sketch.
        sketch.grant_permission(permission='read', user=current_user)
        sketch.grant_permission(permission='write', user=current_user)
        sketch.grant_permission(permission='delete', user=current_user)
        return self.to_json(sketch, status_code=HTTP_STATUS_CODE_CREATED)


class SketchResource(resources.ResourceMixin, Resource):
    """Resource to get a sketch."""

    def _add_label(self, sketch, label):
        """Add a label to the sketch."""
        if sketch.has_label(label):
            logger.warning(
                'Unable to apply the label [{0:s}] to sketch {1:d}, '
                'already exists.'.format(label, sketch.id))
            return False
        sketch.add_label(label, user=current_user)
        return True

    def _remove_label(self, sketch, label):
        """Removes a label to the sketch."""
        if not sketch.has_label(label):
            logger.warning(
                'Unable to remove the label [{0:s}] to sketch {1:d}, '
                'label does not exist.'.format(label, sketch.id))
            return False
        sketch.remove_label(label)
        return True

    def _get_sketch_for_admin(self, sketch):
        """Returns a limited sketch view for adminstrators.

        An administrator needs to get information about all sketches
        that are stored on the backend. However that view should be
        limited for sketches that user does not have explicit read
        or other permissions as well. In those cases the returned
        sketch only contains information about the name, description,
        etc but not any information about the data, nor any access
        to the underlying data of the sketch.

        Args:
            sketch: a sketch object (instance of models.Sketch)

        Returns:
            A limited view of a sketch in JSON (instance of
            flask.wrappers.Response)
        """
        if sketch.get_status.status == 'archived':
            status = 'archived'
        else:
            status = 'admin_view'

        sketch_fields = {
            'id': sketch.id,
            'name': sketch.name,
            'description': sketch.description,
            'user': {'username': current_user.username},
            'timelines': [],
            'stories': [],
            'aggregations': [],
            'aggregationgroups': [],
            'active_timelines': [],
            'label_string': sketch.label_string,
            'status': [{
                'id': 0,
                'status': status}],
            'all_permissions': sketch.all_permissions,
            'created_at': sketch.created_at,
            'updated_at': sketch.updated_at,
        }

        meta = {'current_user': current_user.username}
        return jsonify(
            {
                'objects': [sketch_fields],
                'meta': meta
            }
        )

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.

        Returns:
            A sketch in JSON (instance of flask.wrappers.Response)
        """
        if current_user.admin:
            sketch = Sketch.query.get(sketch_id)
            if not sketch.has_permission(current_user, 'read'):
                return self._get_sketch_for_admin(sketch)
        else:
            sketch = Sketch.query.get_with_acl(sketch_id)

        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        aggregators = {}
        for _, cls in aggregator_manager.AggregatorManager.get_aggregators():
            aggregators[cls.NAME] = {
                'form_fields': cls.FORM_FIELDS,
                'display_name': cls.DISPLAY_NAME,
                'description': cls.DESCRIPTION
            }

        # Get mappings for all indices in the sketch. This is used to set
        # columns shown in the event list.
        sketch_indices = [
            t.searchindex.index_name
            for t in sketch.active_timelines
            if t.get_status.status != 'archived'
        ]

        # Get event count and size on disk for each index in the sketch.
        stats_per_index = {}
        for timeline in sketch.active_timelines:
            if timeline.get_status.status != 'archived':
                continue
            stats_per_index[timeline.searchindex.index_name] = {
                'count': 0,
                'bytes': 0,
                'data_types': []
            }

        if sketch_indices:
            try:
                es_stats = self.datastore.client.indices.stats(
                    index=sketch_indices, metric='docs, store')
            except elasticsearch.NotFoundError:
                es_stats = {}
                logger.error(
                    'Unable to find index in datastore', exc_info=True)

            # Stats for index. Num docs per shard and size on disk.
            for index_name, stats in es_stats.get('indices', {}).items():
                doc_count_all_shards = stats.get(
                    'total', {}).get('docs', {}).get('count', 0)
                bytes_on_disk = stats.get(
                    'total', {}).get('store', {}).get('size_in_bytes', 0)
                num_shards = stats.get('_shards', {}).get('total', 1)
                doc_count = int(doc_count_all_shards / num_shards)

                stats_per_index[index_name] = {
                    'count': doc_count,
                    'bytes': bytes_on_disk
                }

                # Stats per data type in the index.
                parameters = {
                    'limit': '100',
                    'field': 'data_type'
                }
                result_obj, _ = utils.run_aggregator(
                    sketch.id, aggregator_name='field_bucket',
                    aggregator_parameters=parameters,
                    index=[index_name])
                stats_per_index[index_name]['data_types'] = result_obj.values

        if not sketch_indices:
            mappings_settings = {}
        else:
            try:
                mappings_settings = self.datastore.client.indices.get_mapping(
                    index=sketch_indices)
            except elasticsearch.NotFoundError:
                logger.error(
                    'Unable to get indices mapping in datastore', exc_info=True)
                mappings_settings = {}

        mappings = []

        for _, value in mappings_settings.items():
            # The structure is different in ES version 6.x and lower. This check
            # makes sure we support both old and new versions.
            properties = value['mappings'].get('properties')
            if not properties:
                properties = next(
                    iter(value['mappings'].values())).get('properties')

            for field, value_dict in properties.items():
                mapping_dict = {}
                # Exclude internal fields
                if field.startswith('__'):
                    continue
                if field == 'timesketch_label':
                    continue
                mapping_dict['field'] = field
                mapping_dict['type'] = value_dict.get('type', 'n/a')
                mappings.append(mapping_dict)

        # Make the list of dicts unique
        mappings = {v['field']: v for v in mappings}.values()

        views = []
        for view in sketch.get_named_views:
            if not view.user:
                username = 'System'
            else:
                username = view.user.username
            view = {
                'name': view.name,
                'description': view.description,
                'id': view.id,
                'query': view.query_string,
                'filter': view.query_filter,
                'user': username,
                'created_at': view.created_at,
                'updated_at': view.updated_at
            }
            views.append(view)

        meta = dict(
            aggregators=aggregators,
            views=views,
            searchtemplates=[{
                'name': searchtemplate.name,
                'id': searchtemplate.id
            } for searchtemplate in SearchTemplate.query.all()],
            emojis=get_emojis_as_dict(),
            permissions={
                'public': bool(sketch.is_public),
                'read': bool(sketch.has_permission(current_user, 'read')),
                'write': bool(sketch.has_permission(current_user, 'write')),
                'delete': bool(sketch.has_permission(current_user, 'delete')),
            },
            collaborators={
                'users': [user.username for user in sketch.collaborators],
                'groups': [group.name for group in sketch.groups],
            },
            analyzers=[
                x for x, y in analyzer_manager.AnalysisManager.get_analyzers()
            ],
            attributes=utils.get_sketch_attributes(sketch),
            mappings=list(mappings),
            stats=stats_per_index,
            filter_labels=self.datastore.get_filter_labels(
                sketch.id, sketch_indices),
            sketch_labels=[label.label for label in sketch.labels]
        )
        return self.to_json(sketch, meta=meta)

    @login_required
    def delete(self, sketch_id):
        """Handles DELETE request to the resource."""
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'delete'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN, (
                    'User does not have sufficient access rights to '
                    'delete a sketch.'))
        not_delete_labels = current_app.config.get(
            'LABELS_TO_PREVENT_DELETION', [])
        for label in not_delete_labels:
            if sketch.has_label(label):
                abort(
                    HTTP_STATUS_CODE_FORBIDDEN,
                    'Sketch with the label [{0:s}] cannot be deleted.'.format(
                        label))
        sketch.set_status(status='deleted')
        return HTTP_STATUS_CODE_OK

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Returns:
            A sketch in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'write'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have write access controls on sketch.')

        form = request.json
        if not form:
            form = request.data

        update_object = False
        name = form.get('name', '')
        if name:
            sketch.name = name
            update_object = True

        description = form.get('description', '')
        if description:
            sketch.description = description
            update_object = True

        labels = form.get('labels', [])
        label_action = form.get('label_action', 'add')
        if label_action not in ('add', 'remove'):
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Label actions needs to be either "add" or "remove", '
                'not [{0:s}]'.format(label_action))

        if labels and isinstance(labels, (tuple, list)):
            for label in labels:
                if label_action == 'add':
                    changed = self._add_label(sketch=sketch, label=label)
                elif label_action == 'remove':
                    changed = self._remove_label(sketch=sketch, label=label)

                if changed:
                    update_object = True

        if update_object:
            db_session.add(sketch)
            db_session.commit()

        return self.to_json(sketch, status_code=HTTP_STATUS_CODE_CREATED)



"""Tests for SequenceSessionizerSketchPlugin."""

from __future__ import unicode_literals

import mock

from timesketch.lib.analyzers.psexec_sessionizers \
    import DestPsexecSessionizerSketchPlugin
from timesketch.lib.analyzers.sequence_sessionizer \
    import SequenceSessionizerSketchPlugin
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore
from timesketch.lib.analyzers.base_sessionizer_test import _create_eventObj


class ManyEventsSequenceSessionizer(SequenceSessionizerSketchPlugin):
    """Mock sequence sessionizer class with many events in the event_seq."""
    session_type = 'many_events_seq_sessionizer'
    max_time_diff_micros = 100
    return_fields = ['hostname', 'source_short', 'timestamp']
    event_seq = [{
        'hostname': 'host',
        'source_short': 'FILE'
    }, {
        'hostname': 'host',
        'source_short': 'WEBHIST'
    }]


class OneEventSequenceSessionizer(SequenceSessionizerSketchPlugin):
    """Mock sequence sessionizer class with one event in the event_seq."""
    session_type = 'one_event_seq_sessionizer'
    max_time_diff_micros = 100
    event_seq = [{'hostname': 'host', 'source_short': 'FILE'}]
    return_fields = ['hostname', 'source_short', 'timestamp']


# Invalid sequence sessionizers.
class NoneSeqSequenceSessionizer(SequenceSessionizerSketchPlugin):
    """Invalid sequence sessionizer. event_seq should not be None, everything
    else is valid."""
    session_type = 'valid_name'
    event_seq = None
    return_fields = ['timestamp']


class EmptySeqSequenceSessionizer(SequenceSessionizerSketchPlugin):
    """Invalid sequence sessionizer. event_seq should not be [], everything else
    is valid."""
    session_type = 'valid_name'
    event_seq = []
    return_fields = ['timestamp']


class NoTimestampSequenceSessionizer(SequenceSessionizerSketchPlugin):
    """Invalid sequence sessionizer. return_fields should include 'timestamp',
    everything else is valid."""
    session_type = 'valid_name'
    event_seq = [{'hostname': 'host', 'source_short': 'FILE'}]
    return_fields = ['hostname', 'source_short']


class MissingAttrSequenceSessionizer(SequenceSessionizerSketchPlugin):
    """Invalid sequence sessionizer. return_fields doesn't includes all needed
    attributes, everything else is valid."""
    session_type = 'valid_name'
    event_seq = [{'hostname': 'host', 'source_short': 'FILE'}]
    return_fields = ['timestamp']


class NoneSessionTypeSequenceSessionizer(SequenceSessionizerSketchPlugin):
    """Invalid sequence sessionizer. session_type should not be None, everything
    else is valid."""
    session_type = None
    event_seq = [{'hostname': 'host', 'source_short': 'FILE'}]
    return_fields = ['timestamp', 'hostname', 'source_short']


class EmptyStrSessionTypeSequenceSessionizer(SequenceSessionizerSketchPlugin):
    """Invalid sequence sessionizer. session_type should not be empty string,
    everything else is valid."""
    session_type = ''
    event_seq = [{'hostname': 'host', 'source_short': 'FILE'}]
    return_fields = ['timestamp', 'hostname', 'source_short']


class TestValidSequenceSessionizerPlugin(BaseTest):
    """Tests the validation functionality of the sequence sessionizing sketch
    analyzer."""
    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_event_seq_none(self):
        """Test event_seq is not None."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = NoneSeqSequenceSessionizer(index, sketch_id)

        with self.assertRaises(ValueError):
            sessionizer.run()

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_event_seq_empty(self):
        """Test event_seq is not empty."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = EmptySeqSequenceSessionizer(index, sketch_id)

        with self.assertRaises(ValueError):
            sessionizer.run()

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_no_timestamp(self):
        """Test missing timestamp attribute is added in return_fields.
        The sessionizer should be validated automatically when calling
        sessionizer.run()."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = NoTimestampSequenceSessionizer(index, sketch_id)
        sessionizer.datastore.client = mock.Mock()
        datastore = sessionizer.datastore

        _create_mock_event(datastore, 0, 0, [], [0])

        self.assertNotIn('timestamp', sessionizer.return_fields)
        sessionizer.run()
        self.assertIn('timestamp', sessionizer.return_fields)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_missing_attr(self):
        """Test missing attributes added in return_fields.
        The sessionizer should be validated automatically when calling
        sessionizer.run()."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = MissingAttrSequenceSessionizer(index, sketch_id)
        sessionizer.datastore.client = mock.Mock()
        datastore = sessionizer.datastore

        _create_mock_event(datastore, 0, 0, [], [0])

        for event in sessionizer.event_seq:
            for attr in event:
                self.assertNotIn(attr, sessionizer.return_fields)
        sessionizer.run()
        for event in sessionizer.event_seq:
            for attr in event:
                self.assertIn(attr, sessionizer.return_fields)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_session_type_none(self):
        """Test session_type is not None."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = NoneSessionTypeSequenceSessionizer(index, sketch_id)

        with self.assertRaises(ValueError):
            sessionizer.run()

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_session_type_empty_str(self):
        """Test session_type is not empty string."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = EmptyStrSessionTypeSequenceSessionizer(index, sketch_id)

        with self.assertRaises(ValueError):
            sessionizer.run()


class TestManyEventsSequenceSessionizerPlugin(BaseTest):
    """Tests base functionality of sequence sessionizing sketch analyzers with
    many events in the even_seq which are listed in seq_sessionizer_classes.

    Attributes:
        seq_sessionizer_classes: A list of sequence sessionizer classes to
            test.
    """
    seq_sessionizer_classes = [
        ManyEventsSequenceSessionizer,
        DestPsexecSessionizerSketchPlugin
    ]

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_sessionizer(self):
        """Test basic sequence sessionizer functionality."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            self.assertIsInstance(sessionizer, seq_sessionizer_class)
            self.assertEqual(index, sessionizer.index_name)
            self.assertEqual(sketch_id, sessionizer.sketch.id)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_one_session(self):
        """Test one sequence of events is finded and allocated as a session."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            sessionizer.datastore.client = mock.Mock()
            datastore = sessionizer.datastore

            _create_mock_event(datastore,
                               0,
                               2,
                               seq_sessionizer_class.event_seq,
                               time_diffs=[1])

            message = sessionizer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of {0:s} sessions created: 1'.
                format(sessionizer.session_type))

            # Events that are not part of the sequence but are between
            # significant events from the event sequence considered as a session
            # are part of the significant events' session.
            for i in range(0, 101):
                event = datastore.event_store[str(i)]
                self.assertEqual(
                    event['_source']['session_id'][sessionizer.session_type],
                    1)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_multiple_sessions(self):
        """Test multiple sessions are found and allocated correctly."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            sessionizer.datastore.client = mock.Mock()
            datastore = sessionizer.datastore

            _create_mock_event(datastore,
                               0,
                               4,
                               seq_sessionizer_class.event_seq +
                               seq_sessionizer_class.event_seq,
                               time_diffs=[1, 1, 1])

            message = sessionizer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of {0:s} sessions created: 2'.
                format(sessionizer.session_type))

            for i in range(0, 100):
                event = datastore.event_store[str(i)]
                self.assertEqual(
                    event['_source']['session_id'][sessionizer.session_type],
                    1)
            # Events with id in the range of 101 to 201 are not part of any
            # session.
            for i in range(202, 302):
                event = datastore.event_store[str(i)]
                self.assertEqual(
                    event['_source']['session_id'][sessionizer.session_type],
                    2)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_after_session(self):
        """Test events after the last event of a sequence are not allocated with
        a session number if they are not part from another session."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            sessionizer.datastore.client = mock.Mock()
            datastore = sessionizer.datastore

            _create_mock_event(datastore,
                               0,
                               4,
                               seq_sessionizer_class.event_seq +
                               seq_sessionizer_class.event_seq,
                               time_diffs=[1, 1])

            message = sessionizer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of {0:s} sessions created: 2'.
                format(sessionizer.session_type))

            # Session 1: events with id from 0 to 101,
            # session 2: events with id from 202 to 303.
            for i in range(102, 201):
                event = datastore.event_store[str(i)]
                self.assertNotIn('session_id', event['_source'])

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_edge_time_diff(self):
        """Test events with the edge time difference between them are
        allocated correctly."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            sessionizer.datastore.client = mock.Mock()
            datastore = sessionizer.datastore

            _create_mock_event(
                datastore,
                0,
                2,
                seq_sessionizer_class.event_seq,
                time_diffs=[seq_sessionizer_class.max_time_diff_micros])

            message = sessionizer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of {0:s} sessions created: 1'.
                format(sessionizer.session_type))

            for i in range(0, 101):
                event = datastore.event_store[str(i)]
                self.assertEqual(
                    event['_source']['session_id'][sessionizer.session_type],
                    1)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_above_max_time_diff(self):
        """Test events with max time difference + 1 between them are allocated
        correctly."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            sessionizer.datastore.client = mock.Mock()
            datastore = sessionizer.datastore

            _create_mock_event(
                datastore,
                0,
                2,
                seq_sessionizer_class.event_seq,
                time_diffs=[seq_sessionizer_class.max_time_diff_micros + 1])

            message = sessionizer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of {0:s} sessions created: 0'.
                format(sessionizer.session_type))

            # Events with id 0 and id 101 form the requested sequence, but
            # event with id 100 and 101 have max_time_diff_micros + 1 bewtween
            # them
            for i in range(0, 201):
                event = datastore.event_store[str(i)]
                self.assertNotIn('session_id', event['_source'])

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_zero_events(self):
        """Test the behaviour of the sequence sessionizer when given zero
        events."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            sessionizer.datastore.client = mock.Mock()
            datastore = sessionizer.datastore

            _create_mock_event(datastore, 0, 0, [], [0])

            message = sessionizer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of {0:s} sessions created: 0'.
                format(sessionizer.session_type))


class TestOneEventSequenceSessionizerPlugin(BaseTest):
    """Tests base functionality of sequence sessionizing sketch analyzers with
    one event in the even_seq which are listed in seq_sessionizer_classes.

    Attributes:
        seq_sessionizer_classes: A list of sequence sessionizer classes to
            test.
    """
    seq_sessionizer_classes = [OneEventSequenceSessionizer]

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_sessionizer(self):
        """Test basic sequence sessionizer functionality."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            self.assertIsInstance(sessionizer, seq_sessionizer_class)
            self.assertEqual(index, sessionizer.index_name)
            self.assertEqual(sketch_id, sessionizer.sketch.id)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_one_session(self):
        """Test one sequence of events is finded and allocated as a session."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            sessionizer.datastore.client = mock.Mock()
            datastore = sessionizer.datastore

            _create_mock_event(datastore, 0, 1,
                               seq_sessionizer_class.event_seq)

            message = sessionizer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of {0:s} sessions created: 1'.
                format(sessionizer.session_type))

            # Event with id 0 is the significant for the event_seq event.
            event = datastore.event_store['0']
            self.assertEqual(
                event['_source']['session_id'][sessionizer.session_type], 1)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_multiple_sessions(self):
        """Test multiple sessions are finded and allocated correctly."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            sessionizer.datastore.client = mock.Mock()
            datastore = sessionizer.datastore

            _create_mock_event(datastore,
                               0,
                               2,
                               seq_sessionizer_class.event_seq +
                               seq_sessionizer_class.event_seq,
                               time_diffs=[1, 1, 1])

            message = sessionizer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of {0:s} sessions created: 2'.
                format(sessionizer.session_type))

            # Session 1: events with id 0.
            event = datastore.event_store['0']
            self.assertEqual(
                event['_source']['session_id'][sessionizer.session_type], 1)
            # Session 2: events with id 101.
            event = datastore.event_store['101']
            self.assertEqual(
                event['_source']['session_id'][sessionizer.session_type], 2)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_after_session(self):
        """Test events after the last event of a sequence are not allocated with
        a session number if they are not part from another session."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            sessionizer.datastore.client = mock.Mock()
            datastore = sessionizer.datastore

            _create_mock_event(datastore,
                               0,
                               2,
                               seq_sessionizer_class.event_seq +
                               seq_sessionizer_class.event_seq,
                               time_diffs=[1, 1])

            message = sessionizer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of {0:s} sessions created: 2'.
                format(sessionizer.session_type))

            # Session 1: events with id 0.
            # Session 1: events with id 101.
            for i in range(1, 100):
                event = datastore.event_store[str(i)]
                self.assertNotIn('session_id', event['_source'])

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_zero_events(self):
        """Test the behaviour of the sequence sessionizer when given zero
        events."""
        index = 'test_index'
        sketch_id = 1
        for seq_sessionizer_class in self.seq_sessionizer_classes:
            sessionizer = seq_sessionizer_class(index, sketch_id)
            sessionizer.datastore.client = mock.Mock()
            datastore = sessionizer.datastore

            _create_mock_event(datastore, 0, 0, [], [0])

            message = sessionizer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of {0:s} sessions created: 0'.
                format(sessionizer.session_type))


def _create_mock_event(datastore,
                       event_id,
                       quantity,
                       attributes,
                       time_diffs=None):
    """Loads in the datastore mock events that based on the given arguments.
    Args:
        datastore: An instance of MockDataStore.
        event_id: Desired ID for the Event.
        quantity: The number of Events to be generated.
        attributes: A list of dictionaries with attributes for the generated
            events.
        time_diffs: A list of time differences between the generated Events.
    """

    if not time_diffs:
        time_diffs = [0]
    if quantity < 0:
        quantity = abs(quantity)

    # If the list of time differences is too small to be compatible with the
    # quantity of events, then extend the list with the last value for as many
    # items as necessary.
    if quantity - len(time_diffs) > 0:
        time_diffs.extend([time_diffs[len(time_diffs) - 1]] *
                          (quantity - len(time_diffs)))

    event_timestamp = 1410895419859714

    for i in range(quantity):
        _create_eventObj(datastore, event_id, event_timestamp, attributes[i])
        # Adding extra events after every requested event for better simulation
        # of real timeline data i.e. working with a larger dataset.
        for _ in range(100):
            event_timestamp += 1
            event_id += 1
            _create_eventObj(datastore, event_id, event_timestamp)
        event_timestamp += abs(time_diffs[i])
        event_id += 1
